bidi_dir = 'ltr'
html_lang = 'zh-Hans-CN'

index_title = '索引于 {path}'

data_table_header_directory = '文件夹'
data_table_header_name = '文件名'
data_table_header_last_modified = '最后修改'
data_table_header_size = '大小'

explorer_tooltip = '检视模式 [{}]'
explorer_table = '表格检视'
explorer_gallery = '图示检视'
explorer_gallery2 = '图示检视 (+多媒体)'

tools_tooltip = '工具 [{}]'
tools_preview_on = '启用预览'
tools_preview_off = '停用预览'
tools_select_all = '全选'
tools_deselect_all = '取消全选'
tools_expand_all = '展开所有项目'
tools_filter = '筛选'
tools_filter_clear = '清除筛选'

command_tooltip = '命令 [{}]'
command_mkdir = '新文件夹'
command_mkzip = '新压缩文件'
command_mkfile = '新文件'
command_upload = '上传'
command_uploaddir = '上传文件夹'
command_source = '检视源代码'
command_download = '下载'
command_exec = '本地执行'
command_browse = '本地检视'
command_edit = '编辑'
command_editx = '编辑页面'
command_move = '移动'
command_copy = '复制'
command_link = '建立链接'
command_delete = '删除'

command_move_prompt = '移动至此路径或文件夹:'
command_move_prompt_multi = '移动至此文件夹:'
command_copy_prompt = '复制至此路径或文件夹:'
command_copy_prompt_multi = '复制至此文件夹:'
command_link_prompt = '建立链接于此路径或文件夹:'
command_link_prompt_multi = '建立链接于此文件夹:'

label_selected = '已选择 %count% 项'

previewer_toolbar_title = """\
键盘快捷键:
• [Esc]: 关闭预览。
• [I] 或 [Space]: 切换信息栏。
• [Enter]: 在新分页开启。
• [PageUp] 或 [Alt+←] 或 [←]: 显示上一项。
• [PageDown] 或 [Alt+→] 或 [→]: 显示下一项。
• [Home]: 显示第一项。
• [End]: 显示最后一项。
• [+] 或 [Alt+↑] 或 [↑]: 放大。
• [-] 或 [Alt+↓] 或 [↓]: 缩小。
• [Ctrl+0]: 缩放至符合可视范围。
• [Ctrl+1]: 切换缩放为图片大小与最后缩放大小。
• [Ctrl+←] 或 [Shift+←]: 左移。
• [Ctrl+→] 或 [Shift+→]: 右移。
• [Ctrl+↑] 或 [Shift+↑]: 上移。
• [Ctrl+↓] 或 [Shift+↓]: 下移。
"""
previewer_button_previous = '上一项'
previewer_button_next = '下一项'
previewer_button_infobar = '切换信息栏'
previewer_button_close = '关闭预览'

edit_title = '编辑 {path}'
edit_error_no_javascript = '启用 JavaScript 或升级浏览器以支援 JavaScript，此 HTML 编辑器才能正常运作。'
edit_warning_encoding = '此文件以 {encoding} 解码，保存时将编码为 UTF-8。'

markdown_title = '页面 {path}'
maff_index_title = '页面于 {path}'

button_save = '保存'
button_save_tooltip = '保存 [{}]'
button_exit = '离开'
button_exit_tooltip = '离开 [{}]'


#########################################################################
# scrapbook.cache
#########################################################################

# map.html
cache_index_toggle_all = '全部展开或收合'
cache_index_search_link_title = '搜索'
cache_index_source_link_title = '来源链接'

# search.html
cache_search_title = '%book% :: 搜索'
cache_search_view_in_map = '在索引页中检视'
cache_search_start = '查'
cache_search_help_label = '搜索语法说明'
cache_search_help_desc = """\
• 输入的关键词会在标题、内文、评注中查找。
• 可用半形空白分隔要检索的多个关键词。例如“w3c organization”表示搜索含有“w3c”及“organization”的项目 。
• 可用半形双引号检索完整关键词，若关键词包含半形双引号，可用连续两个半形双引号表示。例如“"Tom ""loves"" Mary."”表示搜索含有“Tom "loves" Mary.”的项目。
• 可用负号排除关键词。例如“-virus”表示搜索不含“virus”的项目。
• 可用“<命令>:<关键词>”指定特殊的检索条件，命令前可加负号表示排除或反向。可用命令如下：
  • mc：之后的关键词皆比对大小写。加负号则反之。例如“mc: CIA FBI -mc: president”表示搜索含有区分大小写的“CIA”、“FBI”及不分大小写的“president”的项目。
  • re：之后的关键词皆视为正则表达式。加负号则反之。例如“re: \\bcolou?r\\b -re: 1+1=2”表示搜索匹配正则表达式“\\bcolou?r\\b”且含有关键词“1+1=2”的项目。
  • id：搜索 ID 等同关键词（或与正则表达式匹配）的项目。多次指定时以“或”连接。例如“id:2020 id:2021”表示搜索 ID 为“2020”或“2021”的项目；“-id:2020”表示搜索 ID 不为“2020”的项目。
  • type：搜索类型等同关键词（或与正则表达式匹配）的项目。多次指定时以“或”连接。可用的类型有“”（网页）、“bookmark”（书签）、“file”（文件）、“note”（笔记）等等。例如“type: type:bookmark”表示搜索类型为网页或书签的项目。
  • title：搜索标题含有关键词的项目。
  • content：搜索全文含有关键词的项目。
  • comment：搜索评注含有关键词的项目。
  • tc：搜索标题或评注含有关键词的项目。
  • tcc：搜索标题、全文或评注含有关键词的项目。
  • index：搜索索引文件含有关键词的项目。
  • source：搜索原始网址含有关键词的项目。
  • icon：搜索图示网址含有关键词的项目。
  • charset：搜索字元集含有关键词的项目。
  • create：搜索建立时间匹配条件的项目。多次指定时以“或”连接。时间格式为 0-17 位数字，后面接负号（可略），再接 0-17 位数字，表示时间范围的起始与结束。两个 17 位数字表示本地时间的年（4 位数）、月（01-12）、日（01-31）、时（00-59）、分（00-59）、秒（00-59）、毫秒（000-999），省略的部分视为补 0，唯结束时间全省略时视为“999...”。例如“create:2014-2015”表示 2014 年到 2015 年，“create:-201309”表示 2013 年九月以前，“create:20110825”表示 2011 年八月 25 日以后。
  • modify：搜索修改时间匹配条件的项目。多次指定时以“或”连接。时间格式同 create。
  • marked：搜索强调标示的项目。
  • locked：搜索锁定的项目。
  • location：搜索含有地理位置信息的项目。
  • file：搜索文件名含有关键词的项目。
  • root：搜索此 ID 项目下的子项目。多次指定时以“或”连接。
  • book：搜索指定剪贴簿（依 ID）中的项目。多次指定时以“或”连接。
  • sort：将搜索结果按指定方式排序，负号表示倒序。可填入 id、title、comment、file、content、source、type、create、modify。例如“sort:id -sort:modify”表示先按 ID 递增排序再按修改时间递减排序。
  • limit：设置搜索结果笔数限制。例如“limit:10”表示只呈现前 10 笔搜索结果，“limit:-20”表示不呈现最后 20 笔搜索结果，“limit:0”或“-limit:”表示移除之前设置的限制。"""
cache_search_result = '找到 %length% 笔结果：'
cache_search_result_named = '(%name%) 找到 %length% 笔结果：'
cache_search_sort_last_created = '最后创建'
cache_search_sort_last_modified = '最后修改'
cache_search_sort_title = '标题排序'
cache_search_sort_id = 'ID 排序'
cache_search_confirm_remote = '加载远程全文缓存可能使用较大的网络流量。要继续吗？'


#########################################################################
# WebScrapBook
#########################################################################

EditorDeleteAnnotationConfirm = '删除这个批注吗？'



import argparse

try:
    # since Python 3.9
    BooleanOptionalAction = argparse.BooleanOptionalAction
except AttributeError:
    class BooleanOptionalAction(argparse.Action):
        def __init__(self,
                     option_strings,
                     dest,
                     default=None,
                     type=None,
                     choices=None,
                     required=False,
                     help=None,
                     metavar=None):

            _option_strings = []
            for option_string in option_strings:
                _option_strings.append(option_string)

                if option_string.startswith('--'):
                    option_string = '--no-' + option_string[2:]
                    _option_strings.append(option_string)

            super().__init__(
                option_strings=_option_strings,
                dest=dest,
                nargs=0,
                default=default,
                type=type,
                choices=choices,
                required=required,
                help=help,
                metavar=metavar)

        def __call__(self, parser, namespace, values, option_string=None):
            if option_string in self.option_strings:
                setattr(namespace, self.dest, not option_string.startswith('--no-'))

        def format_usage(self):
            return ' | '.join(self.option_strings)



bidi_dir = 'ltr'
html_lang = 'es'

index_title = 'Índice de {path}'

data_table_header_directory = 'Directorio'
data_table_header_name = 'Nombre'
data_table_header_last_modified = 'Última modificación'
data_table_header_size = 'Tamaño'

explorer_tooltip = 'Modo de vista [{}]'
explorer_table = 'Vista de tabla'
explorer_gallery = 'Vista de galería'
explorer_gallery2 = 'Vista de galería (+media)'

tools_tooltip = 'Herramientas [{}]'
tools_preview_on = 'Habilitar vista previa'
tools_preview_off = 'Deshabilitar vista previa'
tools_select_all = 'Seleccionar todo'
tools_deselect_all = 'Deseleccionar todo'
tools_expand_all = 'Expandir todo'
tools_filter = 'Filtrar'
tools_filter_clear = 'Borrar filtro'

command_tooltip = 'Comandos [{}]'
command_mkdir = 'Nueva carpeta'
command_mkzip = 'Nuevo zip'
command_mkfile = 'Archivo nuevo'
command_upload = 'Cargar'
command_uploaddir = 'Cargar carpeta'
command_source = 'Ver código fuente'
command_download = 'Descargar'
command_exec = 'Ejecutar de forma nativa'
command_browse = 'Navegar de forma nativa'
command_edit = 'Editar'
command_editx = 'Editar página'
command_move = 'Mover'
command_copy = 'Copiar'
command_link = 'Crear enlace'
command_delete = 'Borrar'

command_move_prompt = 'Mover a la ruta o carpeta:'
command_move_prompt_multi = 'Mover a la carpeta:'
command_copy_prompt = 'Copiar a la ruta o carpeta:'
command_copy_prompt_multi = 'Copiar a la carpeta:'
command_link_prompt = 'Create a link at the path or under the folder:****'
command_link_prompt_multi = 'Create links under the folder:*******'

label_selected = '%count% seleccionado'

previewer_toolbar_title = """\
Atajos de teclado:
• [Esc]: cerrar vista previa.
• [I] o [Espacio]: alterna la barra de información.
• [Entrar]: abrir en una nueva pestaña.
• [RePág] o [Alt+←] o [←]: muestra la entrada anterior.
• [PageDown] o [Alt+→] o [→]: muestra la entrada siguiente.
• [Inicio]: muestra la primera entrada.
• [Fin]: muestra la última entrada.
• [+] o [Alt+↑] o [↑]: acercar.
• [-] o [Alt+↓] o [↓]: alejar.
• [Ctrl+0]: vista de ajuste de zoom.
• [Ctrl+1]: zoom entre el tamaño natural y el último tamaño de zoom.
• [Ctrl+←] o [Shift+←]: mover a la izquierda.
• [Ctrl+→] o [Shift+→]: mover a la derecha.
• [Ctrl+↑] o [Shift+↑]: subir.
• [Ctrl+↓] o [Shift+↓]: mover hacia abajo.
"""
previewer_button_previous = 'Anterior'
previewer_button_next = 'Siguiente'
previewer_button_infobar = 'Alternar barra de información'
previewer_button_close = 'Cerrar vista previa'

edit_title = 'Editar {path}'
edit_error_no_javascript = 'Habilite JavaScript o actualice su navegador para admitir JavaScript, de modo que el editor HTML pueda funcionar.'
edit_warning_encoding = 'Este archivo se decodifica como {encoding} y se codificará como UTF-8 cuando se guarde.'

markdown_title = 'Página {path}'
maff_index_title = 'Paginas de {path}'

button_save = 'GUARDAR'
button_save_tooltip = 'Guardar [{}]'
button_exit = 'SALIR'
button_exit_tooltip = 'Salir [{}]'


#########################################################################
# scrapbook.cache
#########################################################################

# map.html
cache_index_toggle_all = 'Alternar todo'
cache_index_search_link_title = 'Búsqueda'
cache_index_source_link_title = 'Enlace de origen'

# search.html
cache_search_title = '%book% :: Búsqueda'
cache_search_view_in_map = 'Ver en mapa'
cache_search_start = 'ir'
cache_search_help_label = 'Ayuda de sintaxis de búsqueda'
cache_search_help_desc = """\
• Búsqueda de palabras clave de entrada de título, contenido o comentario.
• Use espacio para separar varias palabras clave. Por ejemplo, “organización w3c” significa elementos que contienen “w3c” y “organización”.
• Use comillas dobles para marcar una frase completa, una comilla doble literal se puede escapar mediante la duplicación. Por ejemplo, “"Tom ""loves"" Mary."” significa elementos que contienen “Tom "loves" Mary.”
• Utilice el signo menos para excluir una palabra clave. Por ejemplo, “-virus” significa elementos sin “virus”.
• Utilice “<comando>:<palabra clave>” para especificar una condición de búsqueda especial. Se puede anteponer un signo menos para exclusión o reversión. Los comandos disponibles incluyen:
   • mc: cada palabra clave subsiguiente coincide con distinción entre mayúsculas y minúsculas. Por ejemplo, “mc: CIA FBI -mc: presidente” significa que los elementos que contienen “CIA” y “FBI” distinguen mayúsculas de minúsculas y “presidente” distingue entre mayúsculas y minúsculas.
   • re: cada palabra clave posterior se trata como una expresión regular. Por ejemplo, “re: \\bcolou?r\\b -re: 1+1=2” significa elementos que coinciden con la expresión regular “\\bcolou?r\\b” y contienen la palabra clave “1+1=2”.
   • id: elemento cuyo ID es igual a la palabra clave (o coincide con la expresión regular). Múltiples valores están conectados “or". Por ejemplo, “id:2020 id:2021” significa elementos de ID “2020” o “2021”; “-id:2020” significa artículos cuyo ID no es “2020”.
   • type: elementos cuyo tipo es igual a la palabra clave (o coincide con la expresión regular). Múltiples valores están conectados “or”. Los tipos disponibles son “” (página), “bookmark”, “file”, “note”, etc. Por ejemplo, “type: type:bookmark” significa elementos cuyo tipo es página o marcador.
   • title: elementos cuyo título contiene la palabra clave.
   • content: elementos cuyo texto completo contiene la palabra clave.
   • comment: elementos cuyo comentario contiene la palabra clave.
   • tc: artículos cuyo título o comentario contiene la palabra clave.
   • tcc: elementos cuyo título, texto completo o comentario contiene la palabra clave.
   • index: elementos cuya ruta de archivo de índice contiene la palabra clave.
   • source: elementos cuya URL fuente contiene la palabra clave.
   • icon: elementos cuyo icono URL contiene la palabra clave.
   • charset: elementos cuyo juego de caracteres contiene la palabra clave.
   • create: elementos cuyo tiempo de creación coincide con la condición. Múltiples valores están conectados “or”. La condición de tiempo es un intervalo con 0-17 dígitos, seguido de un signo menos opcionalmente, y luego seguido de 0-17 dígitos. Los dos números de 17 dígitos significan el año (4 dígitos), mes (01-12), día (01-31), horas (00-59), minutos (00-59), segundos (00-59) y milisegundos (000-999) en fecha y hora local. Se supone que cada dígito omitido es un "0", excepto que se asume "999..." si se omite totalmente la fecha y hora de finalización. Por ejemplo, “create:2014-2015” significa desde 2014 hasta 2015; “create:-201309” significa antes de septiembre de 2013; y “create:20110825” significa después del 25 de agosto de 2011.
   • modify: elementos cuyo tiempo de modificación coincide con la condición. Múltiples valores están conectados “or”. El formato de hora es el mismo que crear.
   • marked: elementos marcados.
   • locked: elementos bloqueados.
   • location: artículos con información de geolocalización.
   • file: elementos cuyo nombre de archivo contiene la palabra clave.
   • root: elementos bajo el elemento de ID. Múltiples valores están conectados “or”.
   • book: artículos en el álbum de recortes específico (por ID). Múltiples valores están conectados “or”.
   • sort: ordenar los resultados de la búsqueda usando la condición específica, que puede ser id, title, comment, file, content, source, type, create o modify. Por ejemplo, “sort:id -sort:modify” significa ordenar por ID en orden ascendente y luego ordenar por tiempo de modificación en orden descendente.
   • limit: establece un límite en el número de resultados de búsqueda. Por ejemplo, “limit:10” significa mostrar los primeros 10 resultados, “limit:-20” significa eliminar los últimos 20 resultados y “limit:0” o “-limit:” significa anular el límite."""
cache_search_result = 'Encontrado %length% resultados:'
cache_search_result_named = '(%name%) Encontrado %length% resultados:'
cache_search_sort_last_created = 'Última creación'
cache_search_sort_last_modified = 'Última modificación'
cache_search_sort_title = 'Ordenar por título'
cache_search_sort_id = 'Ordenar por ID'


#########################################################################
# WebScrapBook
#########################################################################

EditorDeleteAnnotationConfirm = '¿Eliminar esta anotación?'



"""Search for items in book(s).
"""
import functools
import html
import re
from collections import namedtuple
from contextlib import nullcontext
from datetime import datetime, timedelta, timezone

from .. import util
from .host import Host

Item = namedtuple('Item', ('book_id', 'id', 'file', 'meta', 'fulltext', 'context'))
Sort = namedtuple('Sort', ('key', 'subkey', 'order'), defaults=(None, None, 1))
Date = namedtuple('Date', ('since', 'until'))


class QueryError(Exception):
    pass


class Query:
    """Represents a search query."""
    REPR_FIELDS = ('books', 'roots', 'rules', 'sorts', 'limit', 'mc', 're', 'default')

    PARSE_TEXT_REGEX = re.compile(
        r"""
            (?P<cmd>-*[A-Za-z]+:|-+)
            (?:"(?P<qterm1>[^"]*(?:""[^"]*)*)"|(?P<term1>[^"\s]*))
            |
            (?:"(?P<qterm2>[^"]*(?:""[^"]*)*)"|(?P<term2>[^"\s]+))
        """,
        flags=re.X,
    )

    PARSE_CMD_REGEX = re.compile(r'^(-*)(.*)$')

    PARSE_DATE_REGEX = re.compile(r'^(\d{0,17})(?:-(\d{0,17}))?$')

    TIMEZONE = datetime.now(timezone.utc).astimezone().tzinfo

    ELLIPSIS = '…'

    def __init__(self, query_text):
        """Inatialize a new query.

        Raises:
            ValueError: if any input field cannot be parsed correctly
        """
        self.default = 'tcc'
        self.mc = False
        self.re = False
        self.books = {}
        self.roots = {}
        self.rules = {}
        self.sorts = []
        self.limit = 0

        self.PARSE_TEXT_REGEX.sub(self._parse_query, query_text)
        self.roots.setdefault('include', ['root'])

        self.markers = {
            'title': [
                *self.rules.get('tcc', {}).get('include', []),
                *self.rules.get('tc', {}).get('include', []),
                *self.rules.get('title', {}).get('include', []),
            ],
            'file': [
                *self.rules.get('file', {}).get('include', []),
            ],
            'comment': [
                *self.rules.get('tcc', {}).get('include', []),
                *self.rules.get('tc', {}).get('include', []),
                *self.rules.get('comment', {}).get('include', []),
            ],
            'content': [
                *self.rules.get('tcc', {}).get('include', []),
                *self.rules.get('content', {}).get('include', []),
            ],
            'source': [
                *self.rules.get('source', {}).get('include', []),
            ],
        }

    def __repr__(self):
        cls = self.__class__.__name__
        attrs = ', '.join(f'{f}={getattr(self, f)!r}' for f in self.REPR_FIELDS)
        return f'{cls}({attrs})'

    def _parse_query(self, match):
        cmd, qterm1, term1, qterm2, term2 = match.group('cmd', 'qterm1', 'term1', 'qterm2', 'term2')
        pos = True
        if cmd:
            term = qterm1.replace('""', '"') if qterm1 is not None else term1
            m = self.PARSE_CMD_REGEX.search(cmd)
            if len(m.group(1)) % 2 == 1:
                pos = False
            cmd = m.group(2)
        else:
            term = qterm2.replace('""', '"') if qterm2 is not None else term2
            cmd = ''

        if cmd:
            cmd = cmd[:-1]
        else:
            cmd = self.default

        if cmd == 'default':
            self.default = term
        elif cmd == 'mc':
            self.mc = pos
        elif cmd == 're':
            self.re = pos
        elif cmd == 'book':
            inclusion = 'include' if pos else 'exclude'
            self.books.setdefault(inclusion, []).append(term)
        elif cmd == 'root':
            inclusion = 'include' if pos else 'exclude'
            self.roots.setdefault(inclusion, []).append(term)
        elif cmd == 'sort':
            order = 1 if pos else -1
            if term in ('id', 'file'):
                self.sorts.append(Sort(key=term, order=order))
            elif term == 'content':
                self.sorts.append(Sort(key='fulltext', subkey=term, order=order))
            elif term in ('title', 'comment', 'source', 'type', 'create', 'modify'):
                self.sorts.append(Sort(key='meta', subkey=term, order=order))
            else:
                raise ValueError(f'Invalid sort: {term}')
        elif cmd == 'limit':
            try:
                self.limit = int(term, 10) if pos else 0
            except ValueError:
                raise ValueError(f'Invalid limit: {term}') from None
        elif cmd in ('id', 'type'):
            inclusion = 'include' if pos else 'exclude'
            value = self._parse_str(term, True)
            self.rules.setdefault(cmd, {}).setdefault(inclusion, []).append(value)
        elif cmd in ('file', 'tc', 'tcc', 'title', 'comment', 'content',
                     'index', 'charset', 'source', 'icon'):
            inclusion = 'include' if pos else 'exclude'
            value = self._parse_str(term)
            self.rules.setdefault(cmd, {}).setdefault(inclusion, []).append(value)
        elif cmd in ('create', 'modify'):
            inclusion = 'include' if pos else 'exclude'
            value = self._parse_date(term)
            self.rules.setdefault(cmd, {}).setdefault(inclusion, []).append(value)
        elif cmd in ('marked', 'locked', 'location'):
            inclusion = 'include' if pos else 'exclude'
            self.rules.setdefault(cmd, {}).setdefault(inclusion, True)

    def _parse_str(self, term, exact_match=False):
        flags = (0 if self.mc else re.I) | re.M
        if self.re:
            try:
                return re.compile(term, flags=flags)
            except re.error:
                raise ValueError(f'Invalid regex: {term}') from None
        else:
            key = re.escape(term)
            if exact_match:
                key = '^' + key + '$'
            return re.compile(key, flags=flags)

    @classmethod
    def _parse_date(cls, term):
        match = cls.PARSE_DATE_REGEX.search(term)
        try:
            assert match
            since = cls._parse_date_num(f'{match.group(1):0<17}') if match.group(1) else '0' * 17
            until = cls._parse_date_num(f'{match.group(2):0<17}') if match.group(2) else '9' * 17
        except Exception as exc:
            raise ValueError(f'Invalid date: {term}') from exc
        return Date(since, until)

    @classmethod
    def _parse_date_num(cls, text):
        # Set a slightly narrower upper and lower limit for year to prevent
        # an out-of-range error after applying timedelta and timezone
        # conversion as datetime only supports a year between (1, 9999).
        dt = datetime(
            year=min(max(int(text[0:4], 10), 2), 9998),
            month=min(max(int(text[4:6], 10), 1), 12),
            day=1,
            tzinfo=cls.TIMEZONE,
        )
        # Use a timedelta to prevent an out-of-range error as datetime does not
        # support something like (month=11, day=31).
        delta = timedelta(
            days=max(int(text[6:8], 10), 1) - 1,  # minus starting day 1
            hours=int(text[8:10], 10),
            minutes=int(text[10:12], 10),
            seconds=int(text[12:14], 10),
            milliseconds=int(text[14:17], 10),
        )
        dt = (dt + delta).astimezone(timezone.utc)
        return (f'{dt.year:0>4}{dt.month:0>2}{dt.day:0>2}'
                f'{dt.hour:0>2}{dt.minute:0>2}{dt.second:0>2}'
                f'{(dt.microsecond // 1000):0>3}')

    def match_item(self, item):
        for key, rule in self.rules.items():
            if not getattr(self, f'_match_{key}')(rule, item):
                return False
        return True

    @classmethod
    def _match_tc(cls, rule, item):
        value = '\n'.join([
            item.meta.get('title', ''),
            item.meta.get('comment', ''),
        ])
        return cls.match_text(rule, value)

    @classmethod
    def _match_tcc(cls, rule, item):
        value = '\n'.join([
            item.meta.get('title', ''),
            item.meta.get('comment', ''),
            item.fulltext.get('content', ''),
        ])
        return cls.match_text(rule, value)

    @classmethod
    def _match_content(cls, rule, item):
        value = item.fulltext.get('content')
        return cls.match_text(rule, value)

    @classmethod
    def _match_id(cls, rule, item):
        value = item.id
        return cls.match_text_or(rule, value)

    @classmethod
    def _match_file(cls, rule, item):
        value = item.file
        return cls.match_text(rule, value)

    @classmethod
    def _match_title(cls, rule, item):
        value = item.meta.get('title')
        return cls.match_text(rule, value)

    @classmethod
    def _match_comment(cls, rule, item):
        value = item.meta.get('comment')
        return cls.match_text(rule, value)

    @classmethod
    def _match_index(cls, rule, item):
        value = item.meta.get('index')
        return cls.match_text(rule, value)

    @classmethod
    def _match_charset(cls, rule, item):
        value = item.meta.get('charset')
        return cls.match_text(rule, value)

    @classmethod
    def _match_source(cls, rule, item):
        value = item.meta.get('source')
        return cls.match_text(rule, value)

    @classmethod
    def _match_icon(cls, rule, item):
        value = item.meta.get('icon')
        return cls.match_text(rule, value)

    @classmethod
    def _match_type(cls, rule, item):
        value = item.meta.get('type')
        return cls.match_text_or(rule, value)

    @classmethod
    def _match_create(cls, rule, item):
        value = item.meta.get('create')
        return cls.match_date_or(rule, value)

    @classmethod
    def _match_modify(cls, rule, item):
        value = item.meta.get('modify')
        return cls.match_date_or(rule, value)

    @classmethod
    def _match_marked(cls, rule, item):
        value = item.meta.get('marked')
        return cls.match_bool(rule, value)

    @classmethod
    def _match_locked(cls, rule, item):
        value = item.meta.get('locked')
        return cls.match_bool(rule, value)

    @classmethod
    def _match_location(cls, rule, item):
        value = item.meta.get('location')
        return cls.match_bool(rule, value)

    @staticmethod
    def match_bool(rule, value):
        if rule.get('exclude'):
            if value:
                return False
        if rule.get('include'):
            if not value:
                return False
        return True

    @staticmethod
    def match_text(rule, text):
        text = text or ''
        for key in rule.get('exclude', []):
            if key.search(text):
                return False
        for key in rule.get('include', []):
            if not key.search(text):
                return False
        return True

    @staticmethod
    def match_text_or(rule, text):
        text = text or ''
        for key in rule.get('exclude', []):
            if key.search(text):
                return False
        if not rule.get('include'):
            return True
        for key in rule.get('include', []):
            if key.search(text):
                return True
        return False

    @staticmethod
    def match_date_or(rule, date):
        if not date:
            return False
        for key in rule.get('exclude', []):
            if key[0] <= date <= key[1]:
                return False
        if not rule.get('include'):
            return True
        for key in rule.get('include', []):
            if key[0] <= date <= key[1]:
                return True
        return False

    def get_snippet(self, text, marker_type=None, ln=-1):
        if not text:
            return ''

        regexes = self.markers.get(marker_type, [])
        if ln >= 0:
            if marker_type == 'source':
                text, ellipsis = util.cropped(text, ln, self.ELLIPSIS)
            else:
                text, ellipsis = self._crop_at_first_hit(text, regexes, ln)
        else:
            ellipsis = ''
        return ''.join(self._gen_marked_text(text, regexes)) + self._gen_marked_text_marker(ellipsis)

    @classmethod
    def _crop_at_first_hit(cls, text, regexes, length, context_ratio=0.25):
        min_hit = inf = float('inf')
        for regex in regexes:
            m = regex.search(text)
            if not m:
                continue
            start = m.start(0)
            if start < min_hit:
                min_hit = start
        if min_hit < inf:
            start = max(int(min_hit - length * context_ratio), 0)
            text = text[start:]
        return util.cropped(text, length, cls.ELLIPSIS)

    @classmethod
    def _gen_marked_text(cls, text, regexes):
        ln = len(text)

        hits = []
        for idx, regex in enumerate(regexes):
            pos = 0
            m = regex.search(text, pos)
            while m:
                start, end = m.span(0)
                hit = (start, end, idx)
                hits.append(hit)
                pos = max(hit[1], pos + 1)
                if pos > ln:
                    break
                m = regex.search(text, pos)

        hits = sorted(hits, key=cls._gen_marked_text_sortkey)

        pos = 0
        for hit in hits:
            if hit[0] < pos:
                continue
            delta = text[pos:hit[0]]
            if delta:
                yield cls._gen_marked_text_marker(delta)
            match = text[hit[0]:hit[1]]
            if match:
                yield cls._gen_marked_text_marker(match, hit[2])
            pos = hit[1]
        delta = text[pos:]
        if delta:
            yield cls._gen_marked_text_marker(delta)

    @staticmethod
    def _gen_marked_text_sortkey(hit):
        return hit[0]

    @staticmethod
    def _gen_marked_text_marker(text, idx=False):
        if idx is False:
            return html.escape(text)

        if idx is True:
            return '<mark>' + html.escape(text) + '</mark>'

        return f'<mark class="kw{idx}">' + html.escape(text) + '</mark>'


class SearchEngine:
    def __init__(self, host, query_text, *, lock=True, context=None):
        """Inatialize a new search for the host.

        Raises:
            QueryError: if the query cannot be parsed correctly
        """
        self.host = host
        self.query_text = query_text
        try:
            self.query = Query(query_text)
        except ValueError as exc:
            raise QueryError(str(exc)) from exc
        self.lock = lock
        self.context = context or {}

    def run(self):
        """Start the search and yields result items.

        Yields:
            Item: a found item
        """
        for item in self.search():
            self._generate_context(item)
            yield item

    def search(self):
        results = self.search_books()
        if self.query.limit:
            limit = self.query.limit
            if limit > 0:
                i = 0
                for item in results:
                    i += 1
                    if i > limit:
                        break
                    yield item
            else:
                results = tuple(results)
                for i in range(0, len(results) + limit):
                    yield results[i]
            return
        yield from results

    def search_books(self):
        if self.query.books.setdefault('include', []):
            book_ids = {id: None for id in self.query.books['include'] if id in self.host.books}
        else:
            book_ids = self.host.books

        for book_id in book_ids:
            if book_id in self.query.books.setdefault('exclude', []):
                continue

            lh = self.host.books[book_id].get_tree_lock(persist=self.lock).acquire() if self.lock else nullcontext()
            with lh:
                for item in self.search_book_sorted(book_id):
                    yield item

    def search_book_sorted(self, book_id):
        results = self.search_book(book_id)
        for sort in self.query.sorts:
            keyfunc = functools.partial(self._search_book_sortkey, sort)
            results = sorted(results, key=keyfunc, reverse=sort.order == -1)
        yield from results

    def search_book(self, book_id):
        book = self.host.books[book_id]
        if book.no_tree:
            return

        book.load_meta_files()
        book.load_toc_files()
        book.load_fulltext_files()

        id_pool = {}
        for root in self.query.roots['include']:
            for id in book.get_reachable_items(root):
                id_pool[id] = True
        for root in self.query.roots.setdefault('exclude', []):
            for id in book.get_reachable_items(root):
                try:
                    del id_pool[id]
                except KeyError:
                    pass

        for id in id_pool:
            meta = book.meta.get(id)
            if meta is None:
                continue

            subfiles = book.fulltext.get(id)
            if not subfiles:
                subfiles = {'': {}}

            for file in subfiles:
                item = Item(
                    book_id=book_id,
                    id=id,
                    file=file,
                    meta=meta,
                    fulltext=subfiles[file],
                    context={},
                )
                if self.query.match_item(item):
                    yield item

    @staticmethod
    def _search_book_sortkey(sort, item):
        value = getattr(item, sort.key)
        subkey = sort.subkey
        if subkey is not None:
            value = value.get(sort.subkey, '')
        return value

    def _generate_context(self, item):
        try:
            ln = self.context['title']
            assert isinstance(ln, int)
        except (KeyError, AssertionError):
            pass
        else:
            value = item.meta.get('title', '')
            item.context['title'] = self.query.get_snippet(value, 'title', ln)

        try:
            ln = self.context['file']
            assert isinstance(ln, int)
        except (KeyError, AssertionError):
            pass
        else:
            value = item.file
            item.context['file'] = self.query.get_snippet(value, 'file', ln)

        try:
            ln = self.context['comment']
            assert isinstance(ln, int)
        except (KeyError, AssertionError):
            pass
        else:
            value = item.meta.get('comment', '')
            item.context['comment'] = self.query.get_snippet(value, 'comment', ln)

        try:
            ln = self.context['source']
            assert isinstance(ln, int)
        except (KeyError, AssertionError):
            pass
        else:
            value = item.meta.get('source', '')
            item.context['source'] = self.query.get_snippet(value, 'source', ln)

        try:
            ln = self.context['fulltext']
            assert isinstance(ln, int)
        except (KeyError, AssertionError):
            pass
        else:
            value = item.fulltext.get('content', '')
            item.context['fulltext'] = self.query.get_snippet(value, 'content', ln)


def search(host, query, *, lock=True, context=None):
    """Shorthand to perform a search at given path.

    Raises:
        QueryError: if the query cannot be parsed correctly
    """
    if isinstance(host, Host):
        pass
    elif isinstance(host, str):
        host = Host(host)
    else:
        host = Host(*host)

    engine = SearchEngine(host, query, lock=lock, context=context)
    yield from engine.run()



"""Miscellaneous Scrapbook book handler.
"""
from collections import defaultdict, deque
from contextlib import nullcontext

from . import cache as wsb_cache
from .host import Host


class HostQuery:
    """A utility to perform a series of query on a scrapbook host."""
    def __init__(self, host, query, auto_cache=None, *, lock=True):
        if isinstance(host, Host):
            pass
        elif isinstance(host, str):
            host = Host(host)
        else:
            host = Host(*host)

        self.host = host
        self.query = query
        self.auto_cache = auto_cache
        self.lock = lock

        self.tasks = []
        self.loads = defaultdict(set)
        self.changes = defaultdict(set)
        self.modified = defaultdict(set)
        self.results = []

    def run(self):
        for q in self.query:
            book_id = q.get('book', '')
            cmd = q.get('cmd', '')
            args = q.get('args', ())
            kwargs = q.get('kwargs', {})

            try:
                book = self.host.books[book_id]
            except KeyError:
                raise ValueError(f'Invalid book ID: {book_id!r}') from None

            if book.no_tree:
                raise ValueError(f'Unable to query on a no_tree book: {book_id!r}')

            try:
                func = getattr(book, cmd)
            except AttributeError:
                raise ValueError(f'Invalid command: {cmd!r}') from None

            task = (book_id, cmd, func, args, kwargs)
            self.tasks.append(task)

            changes = getattr(self, f'_cmd_{cmd}_changes', ())
            self.changes[book_id].update(changes)

            loads = getattr(self, f'_cmd_{cmd}_loads', changes)
            self.loads[book_id].update(loads)

            try:
                prehandler = getattr(self, f'_cmd_{cmd}_prehandler')
            except AttributeError:
                pass
            else:
                prehandler(book_id, args, kwargs)

        self.run_tasks()
        return self.results

    def run_tasks(self):
        book_ids = deque(self.changes)
        self._with_next_book(book_ids)

    def _with_next_book(self, book_ids):
        try:
            book_id = book_ids.popleft()
        except IndexError:
            book_id = None

        if book_id is None:
            self._run_tasks()
            return

        book = self.host.books[book_id]
        lh = book.get_tree_lock(persist=self.lock).acquire() if self.lock else nullcontext()
        with lh:
            if 'meta' in self.loads[book_id]:
                book.load_meta_files()

            if 'toc' in self.loads[book_id]:
                book.load_toc_files()

            if 'meta' in self.changes[book_id]:
                book_meta_orig = book.checksum(book.meta)

            if 'toc' in self.changes[book_id]:
                book_toc_orig = book.checksum(book.toc)

            self._with_next_book(book_ids)

            if 'meta' in self.changes[book_id]:
                if book.checksum(book.meta) != book_meta_orig:
                    book.save_meta_files()

            if 'toc' in self.changes[book_id]:
                if book.checksum(book.toc) != book_toc_orig:
                    book.save_toc_files()

    def _run_tasks(self):
        for book_id, cmd, func, args, kwargs in self.tasks:
            try:
                rv = func(*args, **kwargs)
            except Exception as exc:
                task_args = []
                for arg in args:
                    task_args.append(f'{arg!r}')
                for k, v in kwargs.items():
                    task_args.append(f'{k}={v!r}')
                task_args = ', '.join(task_args)
                raise RuntimeError(
                    f'Failed to query: {func.__name__}({task_args}): {exc}'
                ) from exc

            self.results.append(rv)

            try:
                posthandler = getattr(self, f'_cmd_{cmd}_posthandler')
            except (KeyError, AttributeError):
                pass
            else:
                posthandler(book_id, args, kwargs, rv)

        if self.auto_cache:
            # prevent getting an empty set, which means all items
            book_items = {
                book_id: item_ids
                for book_id, item_ids in self.modified.items()
                if item_ids
            }

            if book_items:
                gen = wsb_cache.generate(self.host, book_items,
                                         lock=False, backup=False,
                                         **self.auto_cache)
                for _ in gen:
                    pass

    _cmd_get_item_loads = {'meta', 'toc'}

    _cmd_get_items_loads = _cmd_get_item_loads

    _cmd_add_item_changes = {'meta', 'toc'}

    def _cmd_add_item_posthandler(self, book_id, args, kwargs, rv):
        self.modified[book_id].update(rv)

    _cmd_add_items_changes = _cmd_add_item_changes

    _cmd_add_items_posthandler = _cmd_add_item_posthandler

    _cmd_update_item_changes = {'meta'}

    def _cmd_update_item_posthandler(self, book_id, args, kwargs, rv):
        self.modified[book_id].update(rv)

    _cmd_update_items_changes = _cmd_update_item_changes

    _cmd_update_items_posthandler = _cmd_update_item_posthandler

    _cmd_move_item_loads = {'meta', 'toc'}

    _cmd_move_item_changes = {'toc'}

    _cmd_move_items_loads = _cmd_move_item_loads

    _cmd_move_items_changes = _cmd_move_item_changes

    _cmd_link_item_loads = {'meta', 'toc'}

    _cmd_link_item_changes = {'toc'}

    _cmd_link_items_loads = _cmd_link_item_loads

    _cmd_link_items_changes = _cmd_link_item_changes

    _cmd_copy_item_changes = {'meta', 'toc'}

    def _cmd_copy_item_prehandler(self, book_id, args, kwargs):
        target_book_id = kwargs.get('target_book_id')
        if target_book_id not in (None, book_id):
            self.loads[target_book_id].update({'meta', 'toc'})
            self.changes[target_book_id].update({'meta', 'toc'})

    def _cmd_copy_item_posthandler(self, book_id, args, kwargs, rv):
        _, target_book_id, item_ids = rv
        self.modified[target_book_id].update(item_ids)

    _cmd_copy_items_changes = _cmd_copy_item_changes

    _cmd_copy_items_prehandler = _cmd_copy_item_prehandler

    _cmd_copy_items_posthandler = _cmd_copy_item_posthandler

    _cmd_recycle_item_changes = {'meta', 'toc'}

    _cmd_recycle_items_changes = _cmd_recycle_item_changes

    _cmd_unrecycle_item_changes = {'meta', 'toc'}

    _cmd_unrecycle_items_changes = _cmd_unrecycle_item_changes

    _cmd_delete_item_changes = {'meta', 'toc'}

    def _cmd_delete_item_posthandler(self, book_id, args, kwargs, rv):
        self.modified[book_id].update(rv)

    _cmd_delete_items_changes = _cmd_delete_item_changes

    _cmd_delete_items_posthandler = _cmd_delete_item_posthandler

    _cmd_sort_item_loads = {'meta', 'toc'}

    _cmd_sort_item_changes = {'toc'}

    _cmd_sort_items_loads = _cmd_sort_item_loads

    _cmd_sort_items_changes = _cmd_sort_item_changes

    _cmd_load_item_postit_loads = {'meta'}

    _cmd_save_item_postit_changes = {'meta'}

    def _cmd_save_item_postit_posthandler(self, book_id, args, kwargs, rv):
        self.modified[book_id].update(rv)

    _cmd_add_item_subpage_loads = {'meta'}

    def _cmd_add_item_subpage_posthandler(self, book_id, args, kwargs, rv):
        self.modified[book_id].add(rv)



"""Generator to generate item metadata from files.
"""
import io
import os
import re
import shutil
import traceback
from base64 import b64encode
from datetime import datetime
from functools import partial
from urllib.error import URLError
from urllib.parse import quote, unquote, urljoin, urlsplit, urlunsplit
from urllib.request import pathname2url, url2pathname, urlopen

from lxml import etree

from .. import util
from .._polyfill import mimetypes, zipfile
from ..util import Info
from ..util.css import CssRewriter
from ..util.html import REGEX_ASCII_WHITESPACES, HtmlRewriter

HTML_TITLE_EXCLUDE_PARENTS = {
    'xmp',
    'svg',
    'template',
}

REGEX_IE_DOC_COMMENT = re.compile(r'^\s*saved from url=\((\d+)\)(\S+)\s*')

REGEX_SF_DOC_COMMENT = re.compile(r'^\s+Page saved with SingleFile\s+url: (\S+)\s+saved date: ([^()]+)')

REGEX_MAOXIAN_DOC_COMMENT = re.compile(r'^\s*OriginalSrc: (\S+)')

REGEX_JS_DATE = re.compile(r'^([^()]+)')

COMMON_MIME_EXTENSION = {
    'application/octet-stream': '',
    'text/html': '.html',
    'application/xhtml+xml': '.xhtml',
    'image/svg+xml': '.svg',
    'image/jpeg': '.jpg',
    'audio/mpeg': '.mp3',
    'audio/ogg': '.oga',
    'video/mpeg': '.mpeg',
    'video/mp4': '.mp4',
    'video/ogg': '.ogv',
    'application/ogg': '.ogx',
}

SUPPORT_FOLDER_SUFFIXES = ['.files', '_files']


def generate_item_title(book, id):
    # infer from source
    if book.meta[id].get('source'):
        parts = urlsplit(book.meta[id].get('source'))
        if parts.scheme:
            title = os.path.basename(unquote(parts.path))
            if title:
                return title

    return None


def generate_item_create(book, id):
    # infer from standard timestamp ID
    dt = util.id_to_datetime(id)
    if dt:
        return util.datetime_to_id(dt)

    # infer from ctime of index file
    index = book.meta[id].get('index')
    if index:
        file = os.path.join(book.data_dir, index)
        try:
            ts = os.stat(file).st_ctime
        except OSError:
            pass
        else:
            dt = datetime.fromtimestamp(ts)
            return util.datetime_to_id(dt)

    # infer from modify
    modify = book.meta[id].get('modify')
    if modify:
        return modify


def generate_item_modify(book, id):
    # infer from mtime of index file
    index = book.meta[id].get('index')
    if index:
        file = os.path.join(book.data_dir, index)
        try:
            ts = os.stat(file).st_mtime
        except OSError:
            pass
        else:
            dt = datetime.fromtimestamp(ts)
            return util.datetime_to_id(dt)

    # infer from create (and then ID)
    create = book.meta[id].get('create')
    if create:
        return create


def iter_title_elems(tree):
    """Iterate over valid title elements."""
    def check(elem):
        p = elem.getparent()
        while p is not None:
            if p.tag in HTML_TITLE_EXCLUDE_PARENTS:
                return False
            p = p.getparent()
        return True

    for elem in tree.iter('title'):
        if check(elem):
            yield elem


def iter_favicon_elems(tree):
    """Iterate over valid favicon elements."""
    for elem in tree.iter('link'):
        if 'icon' in elem.attrib.get('rel', '').lower().split():
            yield elem


def mime_to_extension(mime):
    """Fix extension for some common MIME types.

    For example, Python may pick '.xht' instead of '.xhtml' as the primary
    extension for 'application/xhtml+xml'. This translates some common MIME
    types to a fixed mapped extension for better interoperability.
    """
    try:
        return COMMON_MIME_EXTENSION[mime]
    except KeyError:
        return mimetypes.guess_extension(mime) or ''


class Indexer:
    """A class that generates item metadata for files.
    """
    def __init__(self, book, *,
                 handle_ie_meta=True,
                 handle_singlefile_meta=True,
                 handle_savepagewe_meta=True,
                 handle_maoxian_meta=True,
                 ):
        self.book = book
        self.handle_ie_meta = handle_ie_meta
        self.handle_singlefile_meta = handle_singlefile_meta
        self.handle_savepagewe_meta = handle_savepagewe_meta
        self.handle_maoxian_meta = handle_maoxian_meta

    def run(self, files):
        self.book.load_meta_files()

        indexed = {}
        for file in files:
            id = yield from self._index_file(file)
            if id:
                yield Info('info', f'Added item {id!r} for {self.book.get_subpath(file)!r}.')
                indexed[id] = True

        return indexed

    def _index_file(self, file):
        subpath = self.book.get_subpath(file)
        yield Info('debug', f'Indexing {subpath!r}...')

        if not os.path.isfile(file):
            yield Info('error', f'File {subpath!r} does not exist.')
            return None

        _, ext = os.path.splitext(file.lower())
        is_webpage = ext in self.book.ITEM_INDEX_ALLOWED_EXT

        if is_webpage:
            tree = self.book.get_tree_from_index_file(file)
            if tree is None:
                yield Info('error', f'Failed to read document from file {subpath!r}')
                return None

            tree_root = tree.getroot()
            if tree_root is None or tree_root.tag != 'html':
                yield Info('error', f'No html element in file {subpath!r}')
                return None

            html_elem = tree_root

        # prepare meta
        # use empty create and modify to prevent auto-generated by book.add_item()
        meta = {
            'create': '',
            'modify': '',
        }

        if is_webpage:
            # attempt to load metadata generated by certain applications
            if self.handle_ie_meta:
                self._handle_ie_meta(meta, tree_root)

            if self.handle_singlefile_meta:
                self._handle_singlefile_meta(meta, tree_root)

            if self.handle_savepagewe_meta:
                self._handle_savepagewe_meta(meta, tree_root)

            if self.handle_maoxian_meta:
                self._handle_maoxian_meta(meta, tree_root)

            # merge properties from html[data-scrapbook-*] attributes
            for key, value in html_elem.attrib.items():
                if key.startswith('data-scrapbook-'):
                    meta[key[15:]] = value

        # id
        id = meta.get('id')
        if id is not None:
            # if explicitly specified in html attributes, use it or fail out.
            if id in self.book.meta:
                yield Info('error', f'Specified ID {id!r} is already used.')
                return None

            if id in self.book.SPECIAL_ITEM_ID:
                yield Info('error', f'Specified ID {id!r} is invalid.')
                return None

        else:
            # Take base filename as id if it corresponds to standard timestamp
            # format and not used; otherwise generate a new one.
            basepath = os.path.relpath(file, self.book.data_dir)
            basename = os.path.basename(basepath)
            if basename == 'index.html':
                basename = os.path.basename(os.path.dirname(basepath))
            id, _ = os.path.splitext(basename)

            if util.id_to_datetime(id) and id not in self.book.meta:
                meta['id'] = id

        # add to book
        new_items = self.book.add_item(meta, None)
        id, meta = next(iter(new_items.items()))

        # index
        meta['index'] = os.path.relpath(file, self.book.data_dir).replace('\\', '/')

        # type
        if meta.get('type') is None:
            meta['type'] = '' if is_webpage else 'file'

        # title
        if meta.get('title') is None:
            title = None
            if is_webpage:
                title_elem = next(iter_title_elems(tree), None)
                if title_elem is not None:
                    try:
                        title = (
                            (title_elem.text or '')
                            + ''.join(etree.tostring(e, encoding='unicode') for e in title_elem)
                        )
                    except UnicodeDecodeError as exc:
                        yield Info('error', f'Failed to extract title for {id!r}: {exc}', exc=exc)
            if not title or not title.strip():
                title = generate_item_title(self.book, id)
            meta['title'] = title or ''

        # create
        if not meta.get('create'):
            meta['create'] = generate_item_create(self.book, id) or ''

        # modify
        if not meta.get('modify'):
            meta['modify'] = generate_item_modify(self.book, id) or ''

        # icon
        if meta.get('icon') is None:
            if is_webpage:
                try:
                    favicon_elem = next(iter_favicon_elems(tree))
                except StopIteration:
                    pass
                else:
                    meta['icon'] = favicon_elem.attrib.get('href', '')

        generator = FavIconCacher(self.book, cache_archive=True)
        yield from generator.run([id])

        return id

    def _handle_ie_meta(self, meta, root):
        doc_comment = root.getprevious()

        if doc_comment is None:
            return

        if doc_comment.tag != etree.Comment:
            return

        m = REGEX_IE_DOC_COMMENT.search(doc_comment.text)
        if m is None:
            return

        length = m.group(1)
        source = m.group(2)
        try:
            if len(source) == int(length, 10):
                meta['source'] = source
        except ValueError:
            pass

    def _handle_singlefile_meta(self, meta, root):
        try:
            doc_comment = root[0]
        except KeyError:
            return

        if doc_comment.tag != etree.Comment:
            return

        m = REGEX_SF_DOC_COMMENT.search(doc_comment.text)
        if m is None:
            return

        source = m.group(1)
        date_str = m.group(2)
        dt = datetime.strptime(date_str, '%a %b %d %Y %H:%M:%S GMT%z ')

        meta['source'] = source
        meta['create'] = util.datetime_to_id(dt)

    def _handle_savepagewe_meta(self, meta, root):
        node = root.find('.//meta[@name="savepage-url"][@content]')
        if node is not None:
            meta['source'] = node.attrib['content']

        node = root.find('.//meta[@name="savepage-title"][@content]')
        if node is not None:
            meta['title'] = node.attrib['content']

        node = root.find('.//meta[@name="savepage-date"][@content]')
        if node is not None:
            m = REGEX_JS_DATE.match(node.attrib['content'])
            if m:
                dt = datetime.strptime(m.group(1), '%a %b %d %Y %H:%M:%S GMT%z ')
                meta['create'] = util.datetime_to_id(dt)

    def _handle_maoxian_meta(self, meta, root):
        try:
            doc_comment = root[0]
        except KeyError:
            return

        if doc_comment.tag != etree.Comment:
            return

        m = REGEX_MAOXIAN_DOC_COMMENT.search(doc_comment.text)
        if m is None:
            return

        source = m.group(1)

        meta['source'] = source


class FavIconCacher:
    def __init__(self, book, cache_url=True, cache_archive=False, cache_file=False):
        self.book = book
        self.cache_url = cache_url
        self.cache_archive = cache_archive
        self.cache_file = cache_file

        self.favicons = {}
        self.favicon_dir = os.path.normcase(os.path.join(self.book.tree_dir, 'favicon', ''))

    def run(self, item_ids=None):
        self.book.load_meta_files()
        cached = {}

        # handle all items if none specified
        for id in item_ids or self.book.meta:
            if id not in self.book.meta:
                continue

            cache_file = yield from self._cache_favicon(id)
            if cache_file:
                cached[id] = cache_file

        return cached

    def _cache_favicon(self, id):
        yield Info('debug', f'Caching favicon for {id!r}...')

        url = self.book.meta[id].get('icon')
        if not url:
            yield Info('debug', f'Skipped for {id!r}: no favicon to cache.')
            return None

        urlparts = urlsplit(url)

        index = self.book.meta[id].get('index', '')

        # absolute URL
        if urlparts.scheme:
            if self.cache_url:
                return (yield from self._cache_favicon_absolute_url(id, index, url))
            return None

        # skip protocol-relative URL
        if urlparts.netloc:
            return None

        # skip pure query or hash URL
        if not urlparts.path:
            return None

        if util.is_archive(index):
            if self.cache_archive:
                dataurl = yield from self._get_archive_favicon(id, index, url, unquote(urlparts.path))
                if dataurl:
                    return (yield from self._cache_favicon_absolute_url(id, index, dataurl, url))
            return None

        if self.cache_file:
            dataurl = yield from self._get_file_favicon(id, index, url, unquote(urlparts.path))
            if dataurl:
                return (yield from self._cache_favicon_absolute_url(id, index, dataurl, url))

        return None

    def _cache_favicon_absolute_url(self, id, index, url, source_url=None):
        """cache absolute URL (also works for data URL)
        """
        def verify_mime(mime):
            if not mime:
                yield Info('error', f'Unable to cache favicon {util.crop(source_url, 256)!r} for {id!r}: unknown MIME type')
                return False

            if not (mime.startswith('image/') or mime == 'application/octet-stream'):
                yield Info('error', f'Unable to cache favicon {util.crop(source_url, 256)!r} for {id!r}: invalid image MIME type {mime!r}')
                return False

            return True

        def cache_fh(fsrc):
            hash_ = util.checksum(fsrc)
            ext = mime_to_extension(mime)
            fdst = os.path.join(self.book.tree_dir, 'favicon', hash_ + ext)

            if os.path.isfile(fdst):
                yield Info('info', f'Use saved favicon for {util.crop(source_url, 256)!r} for {id!r} at {self.book.get_subpath(fdst)!r}.')
                return fdst

            yield Info('info', f'Saving favicon {util.crop(source_url, 256)!r} for {id!r} at {self.book.get_subpath(fdst)!r}.')
            fsrc.seek(0)
            os.makedirs(os.path.dirname(fdst), exist_ok=True)
            self.book.backup(fdst)
            with open(fdst, 'wb') as fw:
                shutil.copyfileobj(fsrc, fw)
            return fdst

        if source_url is None:
            source_url = url

        try:
            r = urlopen(url)
        except URLError as exc:
            yield Info('error', f'Unable to cache favicon {util.crop(source_url, 256)!r} for {id!r}: unable to fetch favicon URL.', exc=exc)
            return None
        except ValueError as exc:
            yield Info('error', f'Unable to cache favicon {util.crop(source_url, 256)!r} for {id!r}: unsupported or malformatted URL: {exc}', exc=exc)
            return None

        with r as r:
            mime, _ = util.parse_content_type(r.info()['content-type'])
            if not (yield from verify_mime(mime)):
                return None

            cache_file = yield from cache_fh(r)

        self.book.meta[id]['icon'] = util.get_relative_url(
            cache_file,
            os.path.join(self.book.data_dir, os.path.dirname(index)),
            path_is_dir=False,
        )
        return cache_file

    def _get_archive_favicon(self, id, index, url, subpath):
        """Convert in-zip relative favicon path to data URL.
        """
        file = os.path.join(self.book.data_dir, index)

        try:
            if util.is_maff(index):
                page = next(iter(util.get_maff_pages(file)), None)
                if not page:
                    raise RuntimeError('page not found in MAFF')

                refpath = page.indexfilename
                if not refpath:
                    raise RuntimeError('index file not found in MAFF')

                subpath = os.path.dirname(refpath) + '/' + subpath

            try:
                with zipfile.ZipFile(file) as zh:
                    bytes_ = zh.read(subpath)
            except OSError as exc:
                raise RuntimeError(exc.strerror) from exc
        except Exception as exc:
            yield Info('error', f'Failed to read archive favicon {util.crop(url, 256)!r} for {id!r}: {exc}', exc=exc)
            return None

        mime, _ = mimetypes.guess_type(subpath)
        mime = mime or 'application/octet-stream'
        return f'data:{mime};base64,{b64encode(bytes_).decode("ascii")}'

    def _get_file_favicon(self, id, index, url, subpath):
        """Convert relative favicon path to data URL.
        """
        file = os.path.normpath(os.path.join(self.book.data_dir, index, '..', subpath))

        # skip if already in favicon dir
        if os.path.normcase(file).startswith(self.favicon_dir):
            yield Info('debug', f'Skipped favicon {util.crop(url, 256)!r} for {id!r}: already in favicon folder')
            return None

        try:
            with open(file, 'rb') as fh:
                bytes_ = fh.read()
        except OSError as exc:
            yield Info('error', f'Failed to read archive favicon {util.crop(url, 256)!r} for {id!r}: {exc.strerror}', exc=exc)

        mime, _ = mimetypes.guess_type(subpath)
        mime = mime or 'application/octet-stream'
        return f'data:{mime};base64,{b64encode(bytes_).decode("ascii")}'


class SingleHtmlConverter(HtmlRewriter):
    REGEX_SRCSET = re.compile(r"""(\s*)([^ ,][^ ]*[^ ,])(\s*(?: [^ ,]+)?\s*(?:,|$))""")

    def __init__(self, *args, is_svg=None, **kwargs):
        super().__init__(*args, **kwargs)

        self.is_svg = is_svg

        if self.file:
            if is_svg is None:
                self.is_svg = util.is_svg(self.file)

        self.html_converter = SingleHtmlConverter
        self.css_converter = SingleHtmlCssConverter

    def run(self):
        """Overwrite parent to return rewritten HTML instead.
        """
        self.last_html_encoding = None

        markups = self.load(self.file)
        markups = self.rewrite(markups)
        return ''.join(str(m) for m in markups if not m.hidden)

    def rewrite(self, markups):
        def update_context():
            nonlocal context
            context = 'html'
            for tag in stack:
                if tag in {'template', 'svg', 'math'}:
                    context = tag
                    break

        # special handling for SVG
        if self.is_svg:
            return self.rewrite_svg(markups)

        # reset base_url according to the first base[href] element
        for markup in markups:
            if markup.type == 'starttag':
                if markup.tag == 'base':
                    url = markup.getattr('href')
                    if url is not None:
                        self.base_url = urljoin(self.base_url, url)
                    break

        stack = []
        context = 'html'
        i = 0
        while True:
            try:
                markup = markups[i]
            except IndexError:
                break

            if markup.type == 'starttag':
                stack.append(markup.tag)
                update_context()

                if context == 'html':
                    self.rewrite_markup(markup)
                elif context == 'svg':
                    self.rewrite_markup_svg(markup)

            elif markup.type == 'endtag':
                stack.pop()
                update_context()

            elif markup.type == 'data':
                if context in {'html', 'svg'}:
                    # rewrite content of style element
                    try:
                        last_markup = markups[i - 1]
                    except KeyError:
                        pass
                    else:
                        if last_markup.type == 'starttag' and last_markup.tag == 'style':
                            markup.data = self.rewrite_style_text(markup.data)
                            markup.src = None

            i += 1

        return markups

    def rewrite_svg(self, markups):
        i = 0
        while True:
            try:
                markup = markups[i]
            except IndexError:
                break

            if markup.type == 'starttag':
                self.rewrite_markup_svg(markup)

            elif markup.type == 'data':
                # rewrite content of style element
                try:
                    last_markup = markups[i - 1]
                except KeyError:
                    pass
                else:
                    if last_markup.type == 'starttag' and last_markup.tag == 'style':
                        markup.data = self.rewrite_style_text(markup.data)
                        markup.src = None

            i += 1

        return markups

    def rewrite_markup(self, markup):
        self.rewrite_attr(markup, 'style', self.rewrite_style_attr)
        self.rewrite_attr(markup, 'data-scrapbook-shadowdom', self.rewrite_shadowdom_attr)

        if markup.tag == 'meta':
            http_equiv = markup.getattr('http-equiv')
            if http_equiv and http_equiv.lower() == 'refresh':
                self.rewrite_attr(markup, 'content', self.rewrite_meta_refresh_content)
            return

        if markup.tag == 'link':
            rel = markup.getattr('rel')
            if rel:
                rels = REGEX_ASCII_WHITESPACES.split(rel)
                if 'stylesheet' in rels:
                    self.rewrite_attr(markup, 'href', partial(self.rewrite_url, rewrite_css=True))
                elif 'icon' in rels:
                    self.rewrite_attr(markup, 'href', self.rewrite_url)
            return

        if markup.tag in {'body', 'table', 'tr', 'th', 'td'}:
            self.rewrite_attr(markup, 'background', self.rewrite_url)
            return

        if markup.tag == 'frame':
            self.rewrite_attr(markup, 'src', partial(self.rewrite_url, rewrite_doc=True))
            return

        if markup.tag == 'iframe':
            self.rewrite_iframe(markup)
            return

        if markup.tag in {'a', 'area'}:
            self.rewrite_attr(markup, 'href', self.rewrite_url)
            return

        if markup.tag == 'script':
            self.rewrite_attr(markup, 'src', self.rewrite_url)
            return

        if markup.tag == 'img':
            self.rewrite_attr(markup, 'src', self.rewrite_url)
            self.rewrite_attr(markup, 'srcset', self.rewrite_srcset)
            return

        if markup.tag == 'audio':
            self.rewrite_attr(markup, 'src', self.rewrite_url)
            return

        if markup.tag == 'video':
            self.rewrite_attr(markup, 'src', self.rewrite_url)
            self.rewrite_attr(markup, 'poster', self.rewrite_url)
            return

        if markup.tag == 'source':
            self.rewrite_attr(markup, 'src', self.rewrite_url)
            self.rewrite_attr(markup, 'srcset', self.rewrite_srcset)
            return

        if markup.tag == 'track':
            self.rewrite_attr(markup, 'src', self.rewrite_url)
            return

        if markup.tag == 'embed':
            self.rewrite_attr(markup, 'src', self.rewrite_url)
            return

        if markup.tag == 'object':
            self.rewrite_attr(markup, 'data', partial(self.rewrite_url, rewrite_doc=True))
            return

        if markup.tag == 'applet':
            self.rewrite_attr(markup, 'code', self.rewrite_url)
            self.rewrite_attr(markup, 'archive', self.rewrite_url)
            return

        if markup.tag == 'input' and markup.getattr('type', '').lower() == 'image':
            self.rewrite_attr(markup, 'src', self.rewrite_url)
            return

    def rewrite_markup_svg(self, markup):
        self.rewrite_attr(markup, 'style', self.rewrite_style_attr)
        self.rewrite_attr(markup, 'href', self.rewrite_url)
        self.rewrite_attr(markup, 'xlink:href', self.rewrite_url)

    def rewrite_attr(self, markup, target_attr, callback):
        for i, attr_value in enumerate(markup.attrs):
            attr, value = attr_value
            if attr != target_attr:
                continue

            # there's no point to rewrite a boolean attribute
            if value is None:
                continue

            value_new = callback(value)
            if value_new != value:
                markup.attrs[i] = (attr, value_new)
                markup.src = None

    def remove_attr(self, markup, target_attr):
        attrs = []

        for attr_value in markup.attrs:
            attr, value = attr_value
            if attr == target_attr:
                markup.src = None
                continue

            attrs.append((attr, value))

        markup.attrs = attrs

    def rewrite_url(self, url, rewrite_doc=False, rewrite_css=False, meta_refresh=False):
        if meta_refresh:
            udst = urljoin(self.doc_url, url)
        else:
            udst = urljoin(self.base_url, url)

        urlparts = urlsplit(udst)

        # skip non-file URL
        if urlparts.scheme != 'file':
            return url

        fdst = url2pathname(urlparts.path)

        # query/hash only if targeting self
        if os.path.normcase(fdst) == os.path.normcase(url2pathname(urlsplit(self.doc_url).path)):
            return urlunsplit(('', '', '', urlparts.query, urlparts.fragment))

        if meta_refresh:
            return f'urn:scrapbook:convert:skip:url:{url}'

        mime, _ = mimetypes.guess_type(fdst)
        mime = mime or 'application/octet-stream'

        if rewrite_css:
            if udst in self.url_chain:
                return f'urn:scrapbook:convert:circular:url:{url}'

            try:
                conv = self.css_converter(fdst, url_chain=self.url_chain)
                content = conv.run()
                bytes_ = content.encode(conv.encoding)
                return f'data:{mime},{quote(bytes_)}'
            except OSError:
                return f'urn:scrapbook:convert:error:url:{url}'

        if rewrite_doc and util.mime_is_html(mime) or util.mime_is_svg(mime):
            if udst in self.url_chain:
                return f'urn:scrapbook:convert:circular:url:{url}'

            # handle possible meta refresh
            try:
                fdst_mr = util.get_meta_refreshed_file(fdst)
            except util.MetaRefreshError:
                return f'urn:scrapbook:convert:error:url:{url}'

            if fdst_mr:
                fdst = fdst_mr
                udst = urljoin('file:///', pathname2url(fdst))

                if udst in self.url_chain:
                    return f'urn:scrapbook:convert:circular:url:{url}'

                mime, _ = mimetypes.guess_type(fdst)
                mime = mime or 'application/octet-stream'

            if util.mime_is_html(mime) or util.mime_is_svg(mime):
                try:
                    conv = self.html_converter(fdst, url_chain=self.url_chain, parser=self.parser)
                    content = conv.run()
                    bytes_ = content.encode(conv.encoding)
                    self.last_html_encoding = conv.encoding
                    return f'data:{mime},{quote(bytes_)}'
                except OSError:
                    return f'urn:scrapbook:convert:error:url:{url}'

        try:
            with open(fdst, 'rb') as fh:
                bytes_ = fh.read()
        except OSError:
            return f'urn:scrapbook:converter:error:url:{url}'

        if util.is_compressible(mime):
            return f'data:{mime},{quote(bytes_)}'

        return f'data:{mime};base64,{b64encode(bytes_).decode("ascii")}'

    def rewrite_srcdoc(self, text):
        markups = self.loads(text)
        markups = self.rewrite(markups)
        return ''.join(str(m) for m in markups if not m.hidden)

    def rewrite_srcset(self, srcset):
        return self.REGEX_SRCSET.sub(self.rewrite_srcset_sub, srcset)

    def rewrite_srcset_sub(self, m):
        replacement = self.rewrite_url(m.group(2))
        return m.group(1) + replacement + m.group(3)

    def rewrite_style_text(self, text):
        conv = self.css_converter(ref_url=self.base_url, url_chain=self.url_chain)
        return conv.rewrite(
            text,
            rewrite_import_url=partial(conv.rewrite_url, rewrite_css=True),
            rewrite_font_face_url=conv.rewrite_url,
            rewrite_background_url=conv.rewrite_url,
        )

    def rewrite_style_attr(self, text):
        conv = self.css_converter(ref_url=self.base_url, url_chain=self.url_chain)
        return conv.rewrite(
            text,
            rewrite_background_url=conv.rewrite_url,
        )

    def rewrite_meta_refresh_content(self, text):
        time, url, context = util.parse_meta_refresh_content(text)
        if time is not None:
            url = self.rewrite_url(url, meta_refresh=True)
            return f'{time}; url={url}'
        return text

    def rewrite_iframe(self, markup):
        srcdoc = markup.getattr('srcdoc')

        if srcdoc is not None:
            self.rewrite_attr(markup, 'srcdoc', self.rewrite_srcdoc)
            self.remove_attr(markup, 'src')
            return

        for i, attr_value in enumerate(markup.attrs):
            attr, value = attr_value
            if attr != 'src':
                continue

            # there's no point to rewrite a boolean attribute
            if value is None:
                continue

            value_new = self.rewrite_url(value, rewrite_doc=True)
            if value_new != value:
                as_srcdoc = False
                if value_new.startswith('data:'):
                    bytes_, mime, _ = util.parse_datauri(value_new)
                    if util.mime_is_html(mime) and not util.mime_is_xhtml(mime):
                        markup.attrs[i] = ('srcdoc', bytes_.decode(self.last_html_encoding or 'UTF-8'))
                        as_srcdoc = True
                if not as_srcdoc:
                    markup.attrs[i] = (attr, value_new)
                markup.src = None

    def rewrite_shadowdom_attr(self, text):
        markups = self.loads(text)
        markups = self.rewrite(markups)
        return ''.join(str(m) for m in markups if not m.hidden)


class SingleHtmlCssConverter(CssRewriter):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.css_converter = SingleHtmlCssConverter

    def run(self):
        return super().run(
            rewrite_import_url=partial(self.rewrite_url, rewrite_css=True),
            rewrite_font_face_url=self.rewrite_url,
            rewrite_background_url=self.rewrite_url,
        )

    def rewrite_url(self, url, rewrite_css=False):
        udst = urljoin(self.ref_url, url)

        urlparts = urlsplit(udst)

        # skip non-file URL
        if urlparts.scheme != 'file':
            return url

        fdst = url2pathname(urlparts.path)

        # query/hash only if targeting self
        if os.path.normcase(fdst) == os.path.normcase(url2pathname(urlsplit(self.ref_url).path)):
            return urlunsplit(('', '', urlparts.path, urlparts.query, urlparts.fragment))

        mime, _ = mimetypes.guess_type(fdst)
        mime = mime or 'application/octet-stream'

        if rewrite_css:
            if udst in self.url_chain:
                return f'urn:scrapbook:convert:circular:url:{url}'

            try:
                conv = self.css_converter(fdst, url_chain=self.url_chain)
                content = conv.run()
                bytes_ = content.encode(conv.encoding)
                return f'data:{mime},{quote(bytes_)}'
            except OSError:
                return f'urn:scrapbook:convert:error:url:{url}'

        try:
            with open(fdst, 'rb') as fh:
                bytes_ = fh.read()
        except OSError:
            return f'urn:scrapbook:converter:error:url:{url}'

        return f'data:{mime};base64,{b64encode(bytes_).decode("ascii")}'


class UnSingleHtmlConverter(SingleHtmlConverter):
    def __init__(self, *args, ref_file=None, **kwargs):
        super().__init__(*args, **kwargs)

        self.ref_file = ref_file if ref_file is not None else self.file

        # bind ref_file to the constructor for derived conversions to inherit it
        self.html_converter = partial(UnSingleHtmlConverter, ref_file=self.ref_file)
        self.css_converter = partial(UnSingleHtmlCssConverter, ref_file=self.ref_file)

    def rewrite_url(self, url, rewrite_doc=False, rewrite_css=False, meta_refresh=False):
        if meta_refresh:
            udst = urljoin(self.doc_url, url)
        else:
            udst = urljoin(self.base_url, url)

        urlparts = urlsplit(udst)

        # skip non-data URL
        if urlparts.scheme != 'data':
            return url

        try:
            bytes_, mime, params = util.parse_datauri(url)
        except util.DataUriMalformedError:
            traceback.print_exc()
            return f'urn:scrapbook:convert:error:url:{url}'

        bytes_io = io.BytesIO(bytes_)

        if rewrite_css:
            try:
                conv = self.css_converter()
                text = conv.load(bytes_io)
                text = conv.rewrite(
                    text,
                    rewrite_import_url=partial(conv.rewrite_url, rewrite_css=True),
                    rewrite_font_face_url=conv.rewrite_url,
                    rewrite_background_url=conv.rewrite_url,
                )
                bytes_ = text.encode(conv.encoding)
                bytes_io = io.BytesIO(bytes_)
            except Exception:
                traceback.print_exc()
                return f'urn:scrapbook:convert:error:url:{url}'

        elif rewrite_doc and util.mime_is_html(mime) or util.mime_is_svg(mime):
            try:
                encoding = util.fix_codec(params['charset'])
            except KeyError:
                encoding = None
            try:
                conv = self.html_converter(
                    is_xhtml=util.mime_is_xhtml(mime),
                    is_svg=util.mime_is_svg(mime),
                    encoding=encoding, parser=self.parser)
                markups = conv.load(bytes_io)
                markups = conv.rewrite(markups)
                bytes_io = io.BytesIO()
                for markup in markups:
                    if not markup.hidden:
                        bytes_io.write(str(markup).encode(conv.encoding))
            except Exception:
                traceback.print_exc()
                return f'urn:scrapbook:convert:error:url:{url}'

        bytes_io.seek(0)
        sha = util.checksum(bytes_io)
        ext = mime_to_extension(mime)
        basename = f'{sha}{ext}'
        fdst = os.path.join(os.path.dirname(self.ref_file), basename)
        if not os.path.lexists(fdst):
            bytes_io.seek(0)
            with open(fdst, 'wb') as fh:
                shutil.copyfileobj(bytes_io, fh)

        return basename

    def rewrite_iframe(self, markup):
        srcdoc = markup.getattr('srcdoc')

        if srcdoc is not None:
            self.remove_attr(markup, 'src')

            for i, attr_value in enumerate(markup.attrs):
                attr, value = attr_value
                if attr != 'srcdoc':
                    continue

                if value is None:
                    value = ''

                content = self.rewrite_srcdoc(value)
                bytes_ = content.encode('UTF-8')
                dataurl = f'data:text/html;charset=utf-8,{quote(bytes_)}'
                value_new = self.rewrite_url(dataurl)
                markup.attrs[i] = ('src', value_new)
                markup.src = None

            return

        self.rewrite_attr(markup, 'src', partial(self.rewrite_url, rewrite_doc=True))


class UnSingleHtmlCssConverter(SingleHtmlCssConverter):
    def __init__(self, *args, ref_file=None, **kwargs):
        super().__init__(*args, **kwargs)

        self.ref_file = ref_file if ref_file is not None else self.file

        # bind ref_file to the constructor for derived conversions to inherit it
        self.css_converter = partial(UnSingleHtmlCssConverter, ref_file=self.ref_file)

    def rewrite_url(self, url, rewrite_css=False):
        udst = urljoin(self.ref_url, url)

        urlparts = urlsplit(udst)

        # skip non-data URL
        if urlparts.scheme != 'data':
            return url

        try:
            bytes_, mime, params = util.parse_datauri(url)
        except util.DataUriMalformedError:
            traceback.print_exc()
            return f'urn:scrapbook:convert:error:url:{url}'

        bytes_io = io.BytesIO(bytes_)

        if rewrite_css:
            try:
                conv = self.css_converter()
                text = conv.load(bytes_io)
                text = conv.rewrite(
                    text,
                    rewrite_import_url=partial(conv.rewrite_url, rewrite_css=True),
                    rewrite_font_face_url=conv.rewrite_url,
                    rewrite_background_url=conv.rewrite_url,
                )
                bytes_ = text.encode(conv.encoding)
                bytes_io = io.BytesIO(bytes_)
            except Exception:
                traceback.print_exc()
                return f'urn:scrapbook:convert:error:url:{url}'

        bytes_io.seek(0)
        sha = util.checksum(bytes_io)
        ext = mime_to_extension(mime)
        basename = f'{sha}{ext}'
        fdst = os.path.join(os.path.dirname(self.ref_file), basename)
        if not os.path.lexists(fdst):
            bytes_io.seek(0)
            with open(fdst, 'wb') as fh:
                shutil.copyfileobj(bytes_io, fh)

        return basename



import os
import re
import shutil
import time
import traceback
from datetime import datetime
from urllib.parse import unquote
from urllib.request import pathname2url

from lxml import etree

from ... import WSB_CONFIG, WSB_DIR, util
from ...util import Info
from ..host import Host

RDF = '{http://www.w3.org/1999/02/22-rdf-syntax-ns#}'
NS1 = '{http://amb.vis.ne.jp/mozilla/scrapbook-rdf#}'
NS2 = '{scrapbee@163.com}'  # ScrapBee
NC = '{http://home.netscape.com/NC-rdf#}'
RES_PROTOCOL_BASE = 'resource://scrapbook/'
MOZ_ICON_BASE = 'moz-icon://'

REGEX_ID = re.compile(r'\d{14,23}')
REGEX_RDF_ID = re.compile(r'urn:scrapbook:item(\d{14,23})')
REGEX_RDF_SEQ_ID = re.compile(r'urn:scrapbook:(?:item(\d{14,23})|(root))')

LEGACY_TYPE_MAP = {
    'note': 'postit',
    'notex': 'note',
}

PRUNE_FILES = {
    'backup',
    'tree',
    'scrapbook.rdf',
    'cache.rdf',
    'folders.txt',
    'collection.html',
    'combine.html',
    'combine.css',
    'note.html',
    'search.html',
    'sitemap.xsl',
    'note_template.html',
    'notex_template.html',
}


class LegacyBook:
    """Main legacy scrapbook book controller.
    """
    def __init__(self, root):
        self.root = os.path.realpath(root)
        self.meta = {}
        self.toc = {'root': []}
        self.rdf_mtime = 0
        self.sitemaps = {}
        self.modifies = {}

    def load(self):
        yield Info('info', "Loading 'scrapbook.rdf'...")
        yield from self._load_rdf()
        yield Info('info', 'Inspecting data directory...')
        yield from self._load_data_dir()

    def _load_rdf(self):
        rdf_file = os.path.join(self.root, 'scrapbook.rdf')
        try:
            self.rdf_mtime = os.stat(rdf_file).st_mtime
            with open(rdf_file, 'rb') as fh:
                yield from self._parse_rdf(fh)
        except OSError as exc:
            raise RuntimeError(f"Unable to load 'scrapbook.rdf': {exc.strerror}") from exc
        except etree.XMLSyntaxError as exc:
            raise RuntimeError(f"Malformed 'scrapbook.rdf': {exc.args[0]}") from exc

    def _parse_rdf(self, fh):
        for _event, elem in etree.iterparse(
            fh, events=('end',),
            remove_comments=True, encoding='UTF-8',
            tag=(f'{RDF}Description', f'{NC}BookmarkSeparator', f'{RDF}Seq'),
        ):
            yield Info('debug', f'Inspecting element: {elem.tag} ({elem.attrib.get(f"{RDF}about")})')
            if elem.tag in {f'{RDF}Description', f'{NC}BookmarkSeparator'}:
                rid = elem.attrib[f'{RDF}about']
                rid_match = REGEX_RDF_ID.match(rid)
                if not rid_match:
                    continue

                id = rid_match.group(1)
                meta = {'id': id}
                for attr, value in elem.attrib.items():
                    if attr.startswith(NS1) and attr != f'{NS1}id':
                        meta[attr[len(NS1):]] = value
                    elif attr.startswith(NS2) and attr != f'{NS2}id':
                        meta[attr[len(NS2):]] = value
                self.meta.setdefault(id, {}).update(meta)

            elif elem.tag == f'{RDF}Seq':
                rid = elem.attrib[f'{RDF}about']
                rid_match = REGEX_RDF_SEQ_ID.match(rid)
                if not rid_match:
                    continue

                id = rid_match.group(1) or rid_match.group(2)

                for child in elem:
                    if child.tag == f'{RDF}li':
                        ref_rid = child.attrib[f'{RDF}resource']
                        ref_rid_match = REGEX_RDF_ID.match(ref_rid)
                        if not ref_rid_match:
                            continue

                        ref_id = ref_rid_match.group(1)
                        self.toc.setdefault(id, []).append(ref_id)

            # clean up to save memory
            elem.clear()
            while elem.getprevious() is not None:
                del elem.getparent()[0]

    def _load_data_dir(self, load_index_dat=False):
        try:
            dirs = os.scandir(os.path.join(self.root, 'data'))
        except (FileNotFoundError, NotADirectoryError):
            yield Info('debug', 'Skipped inspecting data directory (not found)')
            return
        except OSError as exc:
            raise RuntimeError(f'Failed to scan data directory: {exc.strerror}') from exc

        with dirs as dirs:
            for dir in dirs:
                # check name as ID
                id = dir.name
                yield Info('debug', f'Inspecting item folder {id!r}')
                match_id = REGEX_ID.match(id)
                if not match_id:
                    yield Info('warn', f'Skipped item folder {id!r} (invalid ID)')
                    continue

                # check index.html
                index_file = os.path.join(dir, 'index.html')
                if not os.path.isfile(index_file):
                    yield Info('debug', f"Skipped item folder {id!r} (missing 'index.html')")
                    continue

                # record mtime of index.html
                ts = os.stat(index_file).st_mtime
                mtime = datetime.fromtimestamp(ts)
                modify = util.datetime_to_id(mtime)
                self.modifies[id] = modify

                # record seen sitemap.xml
                sitemap_file = os.path.join(dir, 'sitemap.xml')
                if os.path.isfile(sitemap_file):
                    self.sitemaps[id] = True

                # load index.dat
                if load_index_dat:
                    yield Info('debug', f"Checking 'index.dat' for {id!r}")
                    index_dat_file = os.path.join(dir, 'index.dat')
                    try:
                        assert os.stat(index_dat_file).st_mtime > self.rdf_mtime
                    except (FileNotFoundError, IsADirectoryError, NotADirectoryError, AssertionError):
                        continue
                    except OSError as exc:
                        yield Info('error', f"Failed to read 'index.dat' for {id!r}: {exc.strerror}", exc=exc)
                        continue

                    yield Info('info', f"Reading metadata from newer 'index.dat' for {id!r}")
                    try:
                        with open(index_dat_file, encoding='UTF-8') as fh:
                            meta = {}
                            for line in fh:
                                line = line.rstrip('\n')
                                key, _, value = line.partition('\t')
                                if not value:
                                    continue
                                meta[key] = value
                            self.meta.setdefault(id, {}).update(meta)
                    except OSError as exc:
                        yield Info('error', f"Failed to read 'index.dat' for {id!r}: {exc.strerror}", exc=exc)
                        continue


class Converter:
    def __init__(self, input, output, *, data_files=True, backup=True):
        self.input = os.path.realpath(input)
        self.output = os.path.realpath(output)
        self.data_files = data_files
        self.backup = backup

        self.index_files = {}

    def run(self):
        fsrc = os.path.normpath(os.path.join(__file__, '..', '..', '..', 'resources', 'config.ini'))
        fdst = os.path.normpath(os.path.join(self.output, WSB_DIR, WSB_CONFIG))

        yield Info('info', 'Generating directory...')
        os.makedirs(os.path.dirname(fdst), exist_ok=True)

        if not os.path.isfile(fdst):
            yield Info('info', 'Generating config file...')
            try:
                shutil.copyfile(fsrc, fdst)
            except OSError as exc:
                yield Info('error', f'Unable to generate config file: {exc}', exc=exc)

        yield Info('info', 'Setting up new scrapbook...')
        host = Host(self.output)
        book = host.books['']
        book.load_meta_files()
        book.load_toc_files()

        if self.backup:
            host.init_auto_backup(note='convert-sb2wsb')

        try:
            yield Info('info', 'Loading legacy scrapbook data...')
            book0 = LegacyBook(self.input)
            yield from book0.load()

            yield Info('info', 'Calculating tree data...')
            yield from self._merge_meta(book, book0)
            yield from self._merge_toc(book, book0)

            yield Info('info', 'Copying data files...')
            yield from self._copy_data_files(book, book0)

            if self.data_files:
                yield from self._convert_data_files(book, book0)

            yield Info('info', 'Saving tree files...')
            book.save_meta_files()
            book.save_toc_files()
        finally:
            if self.backup:
                host.init_auto_backup(False)

    def _merge_meta(self, book, book0):
        for id0, meta0 in book0.meta.items():
            yield Info('debug', f'Inspecting item metadata for {id0!r}')
            id = id0
            meta = meta0.copy()

            # meta['type'], meta['marked']
            meta['type'] = LEGACY_TYPE_MAP.get(meta.get('type'), meta.get('type', ''))
            if meta['type'] == 'marked':
                # if sitemap exists, type should be site
                meta['type'] = 'site' if id in book0.sitemaps else ''
                meta['marked'] = True

            # meta['index']
            if meta['type'] not in book.ITEM_TYPES_WITH_OPTIONAL_INDEX:
                meta['index'] = f'{id}/index.html'
            elif meta.get('icon', '').startswith(f'{RES_PROTOCOL_BASE}data/{id}/'):
                # Add a dummy index.html to relate the icon file with the item
                # if it's saved in the item directory. (mainly for ScrapBee)
                meta['index'] = f'{id}/index.html'
                self.index_files[os.path.normpath(os.path.join(book.data_dir, id, 'index.html'))] = True
                yield Info('debug', f'Registering dummy file {meta["index"]!r}')

            # meta['create']
            # fallback to id
            if meta.get('create'):
                meta['create'] = util.datetime_to_id(util.id_to_datetime_legacy(meta['create']))
            else:
                meta['create'] = util.datetime_to_id(util.id_to_datetime_legacy(id))

            # meta['modify']
            # fallback to mtime of index.html and then create (and then id)
            if meta.get('modify'):
                meta['modify'] = util.datetime_to_id(util.id_to_datetime_legacy(meta['modify']))
            else:
                try:
                    meta['modify'] = book0.modifies[id]
                except KeyError:
                    meta['modify'] = meta['create']

            # meta['icon']
            # resolve resource://scrapbook/* and moz-icon://*
            if meta.get('icon', '').startswith(RES_PROTOCOL_BASE):
                if meta.get('index'):
                    src = os.path.normpath(os.path.join(book.data_dir, id))
                else:
                    src = os.path.normpath(book.data_dir)
                dst = os.path.normpath(os.path.join(book.top_dir, unquote(meta['icon'][len(RES_PROTOCOL_BASE):])))
                rel_path = os.path.relpath(dst, src)
                meta['icon'] = pathname2url(rel_path)  # this quotes URL
            elif meta.get('icon', '').startswith(MOZ_ICON_BASE):
                meta['icon-moz'] = meta['icon']
                meta['icon'] = ''

            # meta['comment']
            try:
                if meta['comment']:
                    meta['comment'] = meta['comment'].replace(' __BR__ ', '\n')
            except KeyError:
                pass

            # meta['charset']
            try:
                if meta['chars']:
                    meta['charset'] = meta['chars']
                del meta['chars']
            except KeyError:
                pass

            # meta['lock']
            try:
                if meta['lock']:
                    meta['locked'] = True
                del meta['lock']
            except KeyError:
                pass

            # meta['folder']
            # this should not appear in scrapbook.rdf normally
            try:
                del meta['folder']
            except KeyError:
                pass

            # meta['exported']
            # this should not appear in scrapbook.rdf normally
            try:
                del meta['exported']
            except KeyError:
                pass

            # meta['container']
            # this should not appear in scrapbook.rdf normally
            try:
                del meta['container']
            except KeyError:
                pass

            # add to book
            book.add_item(meta, None)

    def _merge_toc(self, book, book0):
        book.toc.update(book0.toc)
        return
        yield

    def _copy_data_files(self, book, book0):
        with os.scandir(self.input) as dirs:
            for src in dirs:
                if src.name == WSB_DIR:
                    yield Info('warn', f'Skipped copying special directory {WSB_DIR!r}')
                    continue

                if src.name in PRUNE_FILES:
                    if self.backup:
                        yield Info('debug', f'Backup legacy scrapbook entry {src.name!r}')
                        book.backup(src, base=self.input)
                        continue
                    else:
                        yield Info('debug', f'Skipped legacy scrapbook entry {src.name!r}')
                        continue

                dst = os.path.join(book.top_dir, src.name)

                try:
                    shutil.copytree(src, dst)
                except NotADirectoryError:
                    shutil.copy2(src, dst)

        # generate registered dummy index files
        for path in self.index_files:
            if not os.path.lexists(path):
                yield Info('debug', f'Generating registered dummy file {path!r}')
                os.makedirs(os.path.dirname(path), exist_ok=True)
                with open(path, 'wb'):
                    pass

    def _convert_data_files(self, book, book0):
        from .migrate import ConvertDataFilesLegacy
        converter = ConvertDataFilesLegacy(book)
        yield from converter.run()


def run(input, output, *, data_files=True, backup=True):
    start = time.time()
    yield Info('info', 'conversion mode: ScrapBook --> WebScrapBook')
    yield Info('info', f'input directory: {os.path.abspath(input)}')
    yield Info('info', f'output directory: {os.path.abspath(output)}')
    yield Info('info', f'data-files: {data_files}')
    yield Info('info', f'backup: {backup}')
    yield Info('info', '')

    try:
        conv = Converter(input, output, data_files=data_files, backup=backup)
        yield from conv.run()
    except Exception as exc:
        traceback.print_exc()
        yield Info('critical', str(exc), exc=exc)
        return
    else:
        yield Info('info', 'Done.')

    elapsed = time.time() - start
    yield Info('info', f'Time spent: {elapsed} seconds.')



import json
import os
import time
import traceback
from collections import OrderedDict
from contextlib import nullcontext
from datetime import datetime, timedelta

from .. import util
from .._polyfill import mimetypes, zipfile
from ..util import Info
from .book import _id_now
from .host import Host

EXPORTER_VERSION = 2

SCHEME_ITEM_IDS = 0
SCHEME_ROOT_INDEXES = 1


class Exporter():
    """Main class for generating exports."""
    def __init__(self, output, book, *, scheme=SCHEME_ITEM_IDS, singleton=False, stream=None):
        self.output = output
        self.book = book
        self.scheme = scheme
        self.singleton = singleton
        self.stream = stream

    def run(self, items=None, recursive=False):
        self.book.load_meta_files()
        self.book.load_toc_files()

        self.used_ts = set()
        self.map_id_to_eid = {}

        if isinstance(self.output, zipfile.ZipFile):
            cm = nullcontext(self.output)
        else:
            cm = zipfile.ZipFile(self.output, 'w')

        with cm as self._zh:
            if self.scheme == SCHEME_ITEM_IDS:
                yield from self._export_from_item_ids(items, recursive)
            elif self.scheme == SCHEME_ROOT_INDEXES:
                yield from self._export_from_root_indexes(items, recursive)
            else:
                raise ValueError(f'Unknown items scheme: {self.scheme!r}')

        if self.stream is not None:
            yield Info('debug', 'Streaming...', self.stream.get())

    def _export_from_item_ids(self, items, recursive):
        id_pool = set(self.book.meta)
        if items:
            # add descendant id if recursive mode
            if recursive:
                dict_ = {}
                for id in items:
                    self.book.get_reachable_items(id, dict_)
                items = dict_

            id_pool.intersection_update(items)
        else:
            # exclude recycled items by default
            dict_ = self.book.get_reachable_items(self.book.RECYCLE_ITEM_ID)
            id_pool.difference_update(dict_)

        parent_ids = OrderedDict()
        for root in self.book.SPECIAL_ITEM_ID:
            for i, id in self._enum_child_items(root, parent_ids):
                if id in id_pool:
                    yield from self._export_item(i, id, parent_ids)

    def _export_from_root_indexes(self, items, recursive):
        pool = set()
        for item in items:
            root_id, *indexes = item

            # skip invalid item
            if not indexes:
                continue

            # deduplicate
            key = tuple(item)
            if key in pool:
                continue
            pool.add(key)

            item_id = root_id
            parent_ids = []
            for index in indexes:
                parent_ids.append(item_id)
                item_id = self.book.toc[item_id][index]

            yield from self._export_item(index, item_id, parent_ids)
            if recursive:
                for _index, _item_id in self._enum_child_items2(item_id, parent_ids, indexes):
                    # deduplicate
                    key = (root_id, *indexes)
                    if key in pool:
                        continue
                    pool.add(key)

                    yield from self._export_item(_index, _item_id, parent_ids)

    def _enum_child_items(self, id, parent_ids):
        """Generate descendant items for the item of id."""
        # do not export a circular descendant
        if id in parent_ids:
            return

        parent_ids[id] = True
        for i, child_id in enumerate(self.book.toc.get(id, ())):
            yield i, child_id
            yield from self._enum_child_items(child_id, parent_ids)
        parent_ids.popitem()

    def _enum_child_items2(self, id, parent_ids, indexes):
        """Generate descendant items for the item of id."""
        # do not export a circular descendant
        if id in parent_ids:
            return

        parent_ids.append(id)
        for i, child_id in enumerate(self.book.toc.get(id, ())):
            indexes.append(i)
            yield i, child_id
            yield from self._enum_child_items2(child_id, parent_ids, indexes)
            indexes.pop()
        parent_ids.pop()

    def _export_item(self, pos, id, parent_ids):
        if id in self.map_id_to_eid:
            if self.singleton:
                yield Info('debug', f'Skipped exporting item {id!r} (singleton mode)')
                return

        yield Info('debug', f'Exporting item {id!r}')
        try:
            yield from self._export_item_internal(pos, id, parent_ids)
        except Exception as exc:
            # unexpected error
            traceback.print_exc()
            yield Info('error', f'Failed to export {id!r}: {exc}', exc=exc)

    def _export_item_internal(self, pos, id, parent_ids):
        meta = self.book.meta[id]
        index = meta.get('index', '')

        # generate a unique timestamp as prefix
        ets = _id_now()
        while ets in self.used_ts:
            try:
                dt += timedelta(milliseconds=1)  # noqa: F821
            except UnboundLocalError:
                dt = util.id_to_datetime(ets)
                dt += timedelta(milliseconds=1)
            ets = util.datetime_to_id(dt)
        self.used_ts.add(ets)

        # setup an export id (eid), which is unique and is same among all
        # occurrences of the same id, to the ets of the first occurrence
        eid = self.map_id_to_eid.setdefault(id, ets)

        yield Info('info', f'Exporting {id!r} to {ets!r}')
        parents = [{'id': id, 'title': self.book.meta.get(id, {}).get('title', '')} for id in parent_ids]
        meta_data = {'id': id, **meta}
        export_data = {
            'version': EXPORTER_VERSION,
            'id': eid,
            'timestamp': ets,
            'timezone': int(datetime.now().astimezone().utcoffset().total_seconds()),
            'path': parents,
            'index': pos,
        }

        zh = self._zh

        # add topdir and info files
        zh.writestr(f'{ets}/', '')
        fn = f'{ets}/meta.json'
        zh.writestr(fn, json.dumps(meta_data, ensure_ascii=False, indent=2),
                    **util.fs.zip_compression_params(mimetypes.guess_type(fn)[0]))
        fn = f'{ets}/export.json'
        zh.writestr(fn, json.dumps(export_data, ensure_ascii=False, indent=2),
                    **util.fs.zip_compression_params(mimetypes.guess_type(fn)[0]))
        if self.stream is not None:
            yield Info('debug', 'Streaming...', self.stream.get())

        # include data file(s)
        if index:
            zh.writestr(f'{ets}/data/', '')
            src = os.path.join(self.book.data_dir, os.path.dirname(index) if index.endswith('/index.html') else index)
            yield Info('debug', f'Saving data files for {id!r}: {self.book.get_subpath(src)!r}')
            gen = util.fs.zip_compress(zh, src, f'{ets}/data/{os.path.basename(src)}', stream=self.stream)
            if self.stream is not None:
                for bytes_ in gen:
                    yield Info('debug', 'Streaming...', bytes_)

        # include favicon cache
        iconfile = self.book.get_icon_file(meta)
        if not iconfile:
            return

        favicon_dir = os.path.join(self.book.tree_dir, 'favicon', '')
        if not os.path.normcase(iconfile).startswith(os.path.normcase(favicon_dir)):
            return

        zh.writestr(f'{ets}/favicon/', '')
        gen = util.fs.zip_compress(zh, iconfile, f'{ets}/favicon/{os.path.basename(iconfile)}', stream=self.stream)
        if self.stream is not None:
            for bytes_ in gen:
                yield Info('debug', 'Streaming...', bytes_)


def run(host, output, book_id='', items=None, *,
        scheme=SCHEME_ITEM_IDS, recursive=False, singleton=False, stream=None,
        lock=True):
    start = time.time()

    if isinstance(host, Host):
        pass
    elif isinstance(host, str):
        host = Host(host)
    else:
        host = Host(*host)

    # Fail for invalid book ID
    if book_id not in host.books:
        yield Info('error', f'Invalid book {book_id!r}.')
        return

    yield Info('debug', f'Loading book {book_id!r}.')

    try:
        book = host.books[book_id]

        if book.no_tree:
            yield Info('error', f'Unable to export book {book_id!r} ({book.name!r}) (no_tree).')
            return

        yield Info('info', f'Exporting from book {book_id!r} ({book.name!r}).')
        lh = book.get_tree_lock(persist=lock).acquire() if lock else nullcontext()
        with lh:
            generator = Exporter(output, book, scheme=scheme, singleton=singleton, stream=stream)
            yield from generator.run(items, recursive)

    except Exception as exc:
        traceback.print_exc()
        yield Info('critical', str(exc), exc=exc)
        return
    else:
        yield Info('info', 'Done.')

    yield Info('info', '----------------------------------------------------------------------')

    elapsed = time.time() - start
    yield Info('info', f'Time spent: {elapsed} seconds.')



"""Scrapbook book handler.
"""
import functools
import glob
import hashlib
import html
import json
import os
import re
from datetime import datetime, timedelta, timezone
from urllib.parse import unquote, urlsplit
from urllib.request import pathname2url

from .. import util
from .._polyfill import zipfile

# A shortcut for getting an ID at current time. Also for easier mock testing.
_id_now = functools.partial(util.datetime_to_id, None)


class TreeFileError(ValueError):
    def __init__(self, msg, filename=None):
        self.msg = msg
        self.filename = filename


class TreeFileMalformedError(TreeFileError):
    pass


class TreeFileMalformedWrappingError(TreeFileMalformedError):
    """Malformed wrapping part of a tree file
    """


class TreeFileMalformedJsonError(TreeFileMalformedError):
    """A subclass for JSONDecodeError

    This exception should be raised from a JSONDecodeError, and details
        are accessible through __cause__ attribute.
    """


class Book:
    """Main scrapbook book controller.
    """
    # Escape U+2028 and U+2029 for embedded JSON data used as JavaScript code
    # to prevent script breakage and potential security issue in old browsers
    # not supporting ES2019, as they are not allowed in a string literal.
    # https://stackoverflow.com/questions/16005091/node-js-javascript-stringify
    JSON_TRANSLATER = str.maketrans({'\u2028': '\\u2028', '\u2029': '\\u2029'})

    REGEX_TREE_FILE_WRAPPER = re.compile(r'^(?:/\*.*\*/|[^(])+\(([\s\S]*)\)(?:/\*.*\*/|[\s;])*$')
    REGEX_ITEM_POSTIT = re.compile(r'^.*?<pre>\n?([^<]*(?:<(?!/pre>)[^<]*)*)\n</pre>.*$', re.S)

    # A javascript string >= 256 MiB (UTF-16 chars) causes an error in some
    # older browsers. Split into several smaller JavaScript files to prevent
    # such issue.

    # Split at around 256 K items (a meta item is mostly < 512 bytes)
    SAVE_META_THRESHOLD = 256 * 1024

    # Split at around 4 M entries (a TOC entry is mostly < 32 bytes)
    SAVE_TOC_THRESHOLD = 4 * 1024 * 1024

    # Split at at around 128 MiB
    SAVE_FULLTEXT_THRESHOLD = 128 * 1024 * 1024

    REPR_ATTRS = ('id', 'name', 'top_dir')
    DEFAULT_META = {
        'id': None,
        'index': None,
        'title': None,
        'type': None,
        'create': None,
        'modify': None,
        'source': None,
        'icon': None,
        'comment': None,
        'charset': None,
        'marked': None,
        'locked': None,
        'location': None,
    }
    ROOT_ITEM_ID = 'root'
    HIDDEN_ITEM_ID = 'hidden'
    RECYCLE_ITEM_ID = 'recycle'
    SPECIAL_ITEM_ID = dict.fromkeys((
        ROOT_ITEM_ID,
        HIDDEN_ITEM_ID,
        RECYCLE_ITEM_ID,
    ))
    ITEM_TYPES_WITH_OPTIONAL_INDEX = {
        'folder',
        'separator',
        'bookmark',
    }
    ITEM_INDEX_ALLOWED_EXT = {
        '.html',
        '.htz',
        '.maff',
        '.htm',
    }
    ITEM_POSTIT_FORMATTER = """\
<!DOCTYPE html><html><head>\
<meta charset="UTF-8">\
<meta name="viewport" content="width=device-width">\
<style>pre { white-space: pre-wrap; overflow-wrap: break-word; }</style>\
</head><body><pre>
%POSTIT_CONTENT%
</pre></body></html>"""

    TEMPLATE_DIR = 'templates'
    TEMPLATES = {
        ('note', '.html'): {
            'filename': 'note_template.html',
            'content': """\
<!DOCTYPE html>
<html data-scrapbook-type="note">
<head>
<meta charset="UTF-8">
<meta name="viewport" content="width=device-width,initial-scale=1">
<title data-scrapbook-elem="title">%NOTE_TITLE%</title>
</head>
<body>%NOTE_TITLE%</body>
</html>
""",
        },
        ('note', '.md'): {
            'filename': 'note_template.md',
            'content': """%NOTE_TITLE%""",
        },
    }

    def __init__(self, host, book_id=''):
        self.host = host
        self.config = config = host.config['book'][book_id]
        self.id = book_id
        self.name = config['name']
        self.root = host.root
        self.top_dir = os.path.normpath(os.path.join(host.chroot, config['top_dir']))
        self.data_dir = os.path.normpath(os.path.join(self.top_dir, config['data_dir']))
        self.tree_dir = os.path.normpath(os.path.join(self.top_dir, config['tree_dir']))
        self.no_tree = config['no_tree']

        self.meta = None
        self.toc = None
        self.fulltext = None

    def __repr__(self):
        repr_str = ', '.join(f'{attr}={repr(getattr(self, attr))}' for attr in self.REPR_ATTRS)
        return f'{self.__class__.__name__}({repr_str})'

    @staticmethod
    def checksum(obj, method='sha1'):
        """Get a checksum of an object (by its JSONified string)."""
        m = hashlib.new(method)
        gen = json.JSONEncoder(
            ensure_ascii=False,
            check_circular=False,
            separators=(',', ':'),
        ).iterencode(obj)
        for chunk in gen:
            m.update(chunk.encode('UTF-8'))
        return m.digest()

    def get_subpath(self, file):
        """Get subpath of a file related to root.
        """
        return self.host.get_subpath(file)

    def get_tree_file(self, name, index=0):
        return os.path.join(self.tree_dir, f'{name}{index or ""}.js')

    def iter_tree_files(self, name):
        i = 0
        while True:
            file = self.get_tree_file(name, i)
            if not os.path.exists(file):
                break
            yield file
            i += 1

    def iter_meta_files(self):
        yield from self.iter_tree_files('meta')

    def iter_toc_files(self):
        yield from self.iter_tree_files('toc')

    def iter_fulltext_files(self):
        yield from self.iter_tree_files('fulltext')

    def load_tree_file(self, file):
        """Load a tree file.

        Raises:
            OSError: failed to open or read (unlikely to happen as this is
                usually called via iter_tree_files() and file existence has
                been checked in prior)
            TreeFileMalformedError: file malformed
        """
        with open(file, encoding='UTF-8') as fh:
            text = fh.read()

        # avoid error if file is empty
        if text == '':
            return {}

        m = self.REGEX_TREE_FILE_WRAPPER.search(text)

        if not m:
            raise TreeFileMalformedWrappingError('Malformed tree file wrapping', filename=file)

        try:
            return json.loads(m.group(1))
        except json.decoder.JSONDecodeError as exc:
            raise TreeFileMalformedJsonError(f'Malformed tree file: {exc}', filename=file) from exc

    def load_tree_files(self, name):
        data = {}
        for file in self.iter_tree_files(name):
            data.update(self.load_tree_file(file))

        # remove top-level None values to allow quick clear by appending file
        # e.g. add meta1.js with {'id1': None} to quickly delete 'id1' in meta.js
        for k in tuple(data):
            if data[k] is None:
                del data[k]

        return data

    def load_meta_files(self, refresh=False):
        if refresh or self.meta is None:
            self.meta = self.load_tree_files('meta')

    def load_toc_files(self, refresh=False):
        if refresh or self.toc is None:
            self.toc = self.load_tree_files('toc')

    def load_fulltext_files(self, refresh=False):
        if refresh or self.fulltext is None:
            self.fulltext = self.load_tree_files('fulltext')

    def save_tree_file(self, name, index, gen):
        """Save a tree file.

        Raises:
            OSError: failed to write
        """
        file = self.get_tree_file(name, index)
        self.backup(file)
        with open(file, 'w', encoding='UTF-8', newline='\n') as fh:
            for chunk in gen:
                fh.write(chunk.translate(self.JSON_TRANSLATER))

    def _gen_meta_file(self, data):
        yield '/* Feel free to edit this file, but keep data code valid JSON format. */\n'
        yield 'scrapbook.meta('
        yield from json.JSONEncoder(ensure_ascii=False, indent=2).iterencode(data)
        yield ')'

    def save_meta_files(self):
        """Save to tree/meta#.js
        """
        os.makedirs(os.path.join(self.tree_dir), exist_ok=True)
        i = 0
        size = 1
        meta = {}
        for id in tuple(self.meta):
            if self.meta[id] is None:
                del self.meta[id]
                continue
            meta[id] = self.meta[id]
            size += 1
            if size >= self.SAVE_META_THRESHOLD:
                self.save_tree_file('meta', i, self._gen_meta_file(meta))
                i += 1
                size = 0
                meta = {}

        if size:
            self.save_tree_file('meta', i, self._gen_meta_file(meta))
            i += 1

        # remove unused tree/meta#.js
        while True:
            file = self.get_tree_file('meta', i)
            try:
                self.backup(file)
                os.remove(file)
            except FileNotFoundError:
                break
            i += 1

    def _gen_toc_file(self, data):
        yield '/* Feel free to edit this file, but keep data code valid JSON format. */\n'
        yield 'scrapbook.toc('
        yield from json.JSONEncoder(ensure_ascii=False, indent=2).iterencode(data)
        yield ')'

    def save_toc_files(self):
        """Save to tree/toc#.js
        """
        os.makedirs(os.path.join(self.tree_dir), exist_ok=True)
        i = 0
        size = 1
        toc = {}
        for id in tuple(self.toc):
            if self.toc[id] is None:
                del self.toc[id]
                continue
            toc[id] = self.toc[id]
            size += 1 + len(toc[id])
            if size >= self.SAVE_TOC_THRESHOLD:
                self.save_tree_file('toc', i, self._gen_toc_file(toc))
                i += 1
                size = 0
                toc = {}

        if size:
            self.save_tree_file('toc', i, self._gen_toc_file(toc))
            i += 1

        # remove unused tree/toc#.js
        while True:
            file = self.get_tree_file('toc', i)
            try:
                self.backup(file)
                os.remove(file)
            except FileNotFoundError:
                break
            i += 1

    def _gen_fulltext_file(self, data):
        yield '/* This file is generated by WebScrapBook and is not intended to be edited. */\n'
        yield 'scrapbook.fulltext('
        yield from json.JSONEncoder(ensure_ascii=False, indent=1).iterencode(data)
        yield ')'

    def save_fulltext_file(self, i, data):
        self.save_tree_file('fulltext', i, self._gen_fulltext_file(data))

    def save_fulltext_files(self):
        """Save to tree/fulltext#.js
        """
        os.makedirs(os.path.join(self.tree_dir), exist_ok=True)
        i = 0
        size = 1
        fulltext = {}
        for id in tuple(self.fulltext):
            if self.fulltext[id] is None:
                del self.fulltext[id]
                continue
            fulltext[id] = self.fulltext[id]
            for path in fulltext[id]:
                size += len(fulltext[id][path]['content'])
            if size >= self.SAVE_FULLTEXT_THRESHOLD:
                self.save_tree_file('fulltext', i, self._gen_fulltext_file(fulltext))
                i += 1
                size = 0
                fulltext = {}

        if size:
            self.save_tree_file('fulltext', i, self._gen_fulltext_file(fulltext))
            i += 1

        # remove unused tree/fulltext#.js
        while True:
            file = self.get_tree_file('fulltext', i)
            try:
                self.backup(file)
                os.remove(file)
            except FileNotFoundError:
                break
            i += 1

    def backup(self, file, **kwargs):
        """A shortcut for auto backup.
        """
        self.host.auto_backup(file, **kwargs)

    def get_lock(self, name, *args, **kwargs):
        return self.host.get_lock(f'book-{self.id}-{name}', *args, **kwargs)

    def get_tree_lock(self, *args, **kwargs):
        return self.get_lock('tree', *args, **kwargs)

    def get_index_paths(self, index):
        if util.is_maff(index):
            pages = util.get_maff_pages(os.path.join(self.data_dir, index))
            return [p.indexfilename for p in pages]

        if util.is_htz(index):
            return ['index.html']

        return [os.path.basename(index)]

    def get_tree_from_index_file(self, file):
        try:
            if util.is_html(file):
                with open(file, 'rb') as fh:
                    return util.load_html_tree(fh)

            if util.is_htz(file):
                with zipfile.ZipFile(file) as zh:
                    with zh.open('index.html') as fh:
                        return util.load_html_tree(fh)

            if util.is_maff(file):
                info = next(iter(util.get_maff_pages(file)), None)
                if not info:
                    return None

                with zipfile.ZipFile(file) as zh:
                    with zh.open(info.indexfilename) as fh:
                        return util.load_html_tree(fh)
        except (OSError, zipfile.BadZipFile, KeyError):
            return None

        return None

    def get_icon_file(self, item):
        """Get favicon file path of an item.

        Returns:
            str file path of the favicon, or None if not determinable
        """
        icon = item.get('icon', '')

        if not icon:
            return None

        u = urlsplit(icon)

        if u.scheme:
            return None

        if u.netloc:
            return None

        if not u.path:
            return None

        if u.path.startswith('/'):
            return None

        index = item.get('index', '')
        return os.path.normpath(os.path.join(self.data_dir, os.path.dirname(index), unquote(u.path)))

    def load_postit_file(self, file):
        with open(file, encoding='UTF-8') as fh:
            content = fh.read()

        return self.REGEX_ITEM_POSTIT.sub(r'\1', content)

    def save_postit_file(self, file, content):
        data = util.format_string(self.ITEM_POSTIT_FORMATTER, {
            'POSTIT_CONTENT': content,
        })
        # enforce LF to prevent bad parsing for legacy ScrapBook
        with open(file, 'w', encoding='UTF-8', newline='\n') as fh:
            fh.write(data)

    def get_reachable_items(self, item_id, dict=None):
        """Get a flattened set of reachable items, including self.

        Args:
            item_id: the item id to search from
            dict: a dict to store the final result, or None to generate one
        """
        if dict is None:
            dict = {}

        self._get_reachable_items(item_id, dict)
        return dict

    def _get_reachable_items(self, item_id, dict):
        if item_id in dict:
            return

        dict[item_id] = True

        try:
            child_ids = self.toc[item_id]
        except KeyError:
            return

        for child_id in child_ids:
            self._get_reachable_items(child_id, dict)

    def get_unique_id(self, item_id=None):
        """Get an unique item ID.

        Args:
            item_id: a pre-generated item ID in the format of datetime_to_id()
        """
        if item_id is None:
            item_id = _id_now()

        while (item_id in self.meta
               or os.path.lexists(os.path.join(self.data_dir, item_id))
               or next(glob.iglob(glob.escape(os.path.join(self.data_dir, item_id)) + '.*'), None)
               ):
            try:
                dt += timedelta(milliseconds=1)  # noqa: F821
            except UnboundLocalError:
                dt = util.id_to_datetime(item_id) or datetime.now(timezone.utc)
                dt += timedelta(milliseconds=1)
            item_id = util.datetime_to_id(dt)

        return item_id

    def get_template(self, type, ext='.html'):
        try:
            tpl = self.TEMPLATES[(type, ext)]
        except KeyError:
            raise ValueError(f'Unknown template: {(type, ext)!r}')

        file = os.path.join(self.tree_dir, self.TEMPLATE_DIR, tpl['filename'])

        # generate if not exists
        if not os.path.lexists(file):
            util.fs.save(file, tpl['content'].encode('UTF-8'))

        with open(file, encoding='UTF-8') as fh:
            content = fh.read()

        return content

    def get_item(self, item_id, include_parents=False):
        """Singular version shortcut of get_items()."""
        return self.get_items((item_id,), include_parents).get(item_id)

    def get_items(self, items, include_parents=False):
        """Get information about items in the book.

        Args:
            items: an iterable of item IDs
            include_parents: whether to include parents information

        Returns:
            dict: information of the items
        """
        reverse_toc = {}
        if include_parents:
            for parent_id, toc in self.toc.items():
                for i, item_id in enumerate(toc):
                    reverse_toc.setdefault(item_id, []).append(
                        (parent_id, i),
                    )

        results = {}
        for item_id in items:
            result = {}

            meta = self.meta.get(item_id)
            if meta is not None:
                result['meta'] = meta

            children = self.toc.get(item_id)
            if children is not None:
                result['children'] = children

            if include_parents:
                parents = reverse_toc.get(item_id)
                if parents is not None:
                    result['parents'] = parents

            if result:
                results[item_id] = result

        return results

    def add_item(self, item=None, target_parent_id=ROOT_ITEM_ID, target_index=None):
        """Singular version shortcut of add_items()."""
        return self.add_items((item,), target_parent_id, target_index)

    def add_items(self, items, target_parent_id=ROOT_ITEM_ID, target_index=None):
        """Add items to the book.

        Args:
            items: an iterable of dict with item properties (or None to
                generate a new item).
            target_parent_id: None to not insert to any parent
            index: None to append to last

        Returns:
            dict: ID and meta of the added items

        Raises:
            ValueError: if the provided item ID already exists, the target
                parent does not exist, the target index is invalid, etc.
        """
        if not (target_parent_id is None
                or target_parent_id in self.meta
                or target_parent_id in self.SPECIAL_ITEM_ID):
            raise ValueError(f'Invalid target parent ID: {target_parent_id!r}')

        if not (target_index is None or target_index >= 0):
            raise ValueError(f'Invalid target index: {target_index!r}')

        added = {}
        for item in items:
            if item is None:
                continue

            try:
                item_id = item['id']
            except KeyError:
                continue

            if item_id in self.SPECIAL_ITEM_ID:
                raise ValueError(f'Item ID is preserved: {item_id!r}')

            if item_id in self.meta:
                raise ValueError(f'Item already exists: {item_id!r}')

            if item_id in added:
                raise ValueError(f'Item is duplicated: {item_id!r}')

            added[item_id] = True

        # add to meta (uniquify ID)
        rv = {}
        for item in items:
            item_id = self._add_item(item)
            rv[item_id] = self.meta[item_id]

        # add to TOC if target parent is provided
        if rv and target_parent_id is not None:
            # adjust target_index
            if target_index is None:
                target_index = 0 if self.config['new_at_top'] else float('inf')
            target_index = min(target_index, len(self.toc.get(target_parent_id, ())))

            self.toc.setdefault(target_parent_id, [])[target_index:target_index] = rv

        return rv

    def _add_item(self, meta):
        # prepare new item meta
        item = self.DEFAULT_META.copy()
        if meta is not None:
            item.update(meta)

        if item.get('id') is None:
            # use the same base timestamp for id and create
            item_id = _id_now()
            item['id'] = self.get_unique_id(item_id)
            if item.get('create') is None:
                item['create'] = item_id
        elif item.get('create') is None:
            item['create'] = _id_now()

        if item.get('modify') is None:
            item['modify'] = item['create']

        # remove None keys
        for key in tuple(item):
            if item[key] is None:
                del item[key]

        # add to meta
        self.meta[item['id']] = item

        # remove ID field
        item_id = item.pop('id')

        return item_id

    def update_item(self, item, auto_modify=True):
        """Singular version shortcut of update_items()."""
        return self.update_items((item,), auto_modify=auto_modify)

    def update_items(self, items, auto_modify=True):
        """Update items.

        Args:
            items: an iterable of dict with item properties
            auto_modify: automatically update the 'modify' property (overwrites
                the property value of the provided item)

        Returns:
            dict: ID and meta of the updated items

        Raises:
            ValueError: if the item ID is not specified, the provided item
                does not exist, etc.
        """
        # prepare items to handle
        tasks = {}
        for item in items:
            try:
                item_id = item['id']
            except KeyError:
                raise ValueError(f'Item ID not specified: {item!r}')

            if item_id not in self.meta:
                raise ValueError(f'Item not exist: {item_id!r}')

            # Prevent handling the same item again.
            if item_id in tasks:
                continue

            tasks[item_id] = item

        # perform the tasks
        if auto_modify:
            modify_ts = _id_now()

        rv = {}
        for item_id, item in tasks.items():
            # update meta
            self.meta[item_id].update(item)

            # update modify
            if auto_modify:
                self.meta[item_id]['modify'] = modify_ts

            # remove ID field
            del self.meta[item_id]['id']

            rv[item_id] = self.meta[item_id]

        return rv

    def move_item(self, current_parent_id, current_index,
                  target_parent_id, target_index=None):
        """Singular version shortcut of move_items()."""
        return self.move_items(((current_parent_id, current_index),),
                               target_parent_id, target_index)

    def move_items(self, items, target_parent_id, target_index=None):
        """Move items.

        Args:
            items: an iterable of tuple (current_parent_id, current_index) in
                tree order
            target_parent_id: the parent to insert at
            index: the position to insert at, or None to append to last

        Returns:
            int: index that the items are inserted at

        Raises:
            ValueError: if the item does not exist, the target parent does not
                exist, the target index is invalid, etc.
        """
        if not (target_parent_id in self.meta or target_parent_id in self.SPECIAL_ITEM_ID):
            raise ValueError(f'Invalid target parent ID: {target_parent_id!r}')

        if not (target_index is None or target_index >= 0):
            raise ValueError(f'Invalid target index: {target_index!r}')

        # adjust target_index
        if target_index is None:
            target_index = 0 if self.config['new_at_top'] else float('inf')
        target_index = min(target_index, len(self.toc.get(target_parent_id, ())))

        # prepare items to handle
        tasks = {}
        for current_parent_id, current_index in items:
            if not (current_index >= 0):
                raise ValueError(f'Invalid item index: {current_index!r}')

            try:
                item_id = self.toc[current_parent_id][current_index]
            except (KeyError, IndexError):
                raise ValueError(
                    f'Item not exist: {current_parent_id!r}[{current_index!r}]'
                ) from None

            item = (item_id, current_parent_id, current_index)

            # Prevent handling the same item again, which may happen when the
            # user selects a recursive tree.
            #
            # For example:
            #
            #     root
            #       item1
            #         item2
            #           item1
            #             item2
            #
            # In this case the user may select:
            #
            #     [('root', 0) ('item1', 0), ('item2', 0), ('item1', 0)]
            #
            # and the last item should be ignored.
            #
            if item in tasks:
                continue

            # Silently ignore moving into a descendant as it will become
            # non-reachable (unless move within the same parent).
            if (target_parent_id in self.get_reachable_items(item_id)
                    and current_parent_id != target_parent_id):
                continue

            tasks[item] = True

        # perform the tasks
        if tasks:
            try:
                it = reversed(tasks)
            except TypeError:
                # Python 3.7 does not support dict reverse
                it = reversed(tuple(tasks))

            for _, current_parent_id, current_index in it:
                # remove from parent TOC
                del self.toc[current_parent_id][current_index]
                if not self.toc[current_parent_id]:
                    del self.toc[current_parent_id]

                # fix when moving within the same parent
                if current_parent_id == target_parent_id and current_index < target_index:
                    target_index -= 1

            self.toc.setdefault(target_parent_id, [])[target_index:target_index] = (
                item_id for item_id, _, _ in tasks
            )

        return target_index

    def link_item(self, current_parent_id, current_index,
                  target_parent_id, target_index=None):
        """Singular version shortcut of link_items()."""
        return self.link_items(((current_parent_id, current_index),),
                               target_parent_id, target_index)

    def link_items(self, items, target_parent_id, target_index=None):
        """Create links for items.

        Args:
            items: an iterable of tuple (current_parent_id, current_index) in
                tree order
            target_parent_id: the parent to insert at
            index: the position to insert at, or None to append to last

        Returns:
            int: index that the items are inserted at

        Raises:
            ValueError: if the item does not exist, the target parent does not
                exist, the target index is invalid, etc.
        """
        if not (target_parent_id in self.meta or target_parent_id in self.SPECIAL_ITEM_ID):
            raise ValueError(f'Invalid target parent ID: {target_parent_id!r}')

        if not (target_index is None or target_index >= 0):
            raise ValueError(f'Invalid target index: {target_index!r}')

        # adjust target_index
        if target_index is None:
            target_index = 0 if self.config['new_at_top'] else float('inf')
        target_index = min(target_index, len(self.toc.get(target_parent_id, ())))

        # prepare items to handle
        tasks = {}
        for current_parent_id, current_index in items:
            if not (current_index >= 0):
                raise ValueError(f'Invalid item index: {current_index!r}')

            try:
                item_id = self.toc[current_parent_id][current_index]
            except (KeyError, IndexError):
                raise ValueError(
                    f'Item not exist: {current_parent_id!r}[{current_index!r}]'
                ) from None

            item = (item_id, current_parent_id, current_index)

            # Prevent handling the same item again, as move_items().
            if item in tasks:
                continue

            tasks[item] = True

        # perform the tasks
        if tasks:
            self.toc.setdefault(target_parent_id, [])[target_index:target_index] = (
                item_id for item_id, _, _ in tasks
            )

        return target_index

    def copy_item(self, current_parent_id, current_index,
                  target_parent_id, target_index=None, *,
                  target_book_id=None, recursively=True):
        """Singular version shortcut of copy_items()."""
        return self.copy_items(((current_parent_id, current_index),),
                               target_parent_id, target_index,
                               target_book_id=target_book_id,
                               recursively=recursively,
                               )

    def copy_items(self, items, target_parent_id, target_index=None, *,
                   target_book_id=None, recursively=True):
        """Copy items.

        Args:
            items: an iterable of tuple (current_parent_id, current_index) in
                tree order
            target_parent_id: the parent to insert at
            index: the position to insert at, or None to append to last
            target_book_id: the ID of the scrapbook copy to
            recursively: also copy descendant items

        Returns:
            int: index that the items are inserted at
            str: ID of the target book
            list: ID of the generated items

        Raises:
            ValueError: if the item does not exist, the target parent does not
                exist, the target index is invalid, etc.
        """
        if target_book_id is None:
            target_book_id = self.id

        if target_book_id == self.id:
            target_book = self
        else:
            try:
                target_book = self.host.books[target_book_id]
            except KeyError:
                raise ValueError(f'Invalid target book ID: {target_book_id!r}')

            target_book.load_meta_files()
            target_book.load_toc_files()

        if not (target_parent_id in target_book.meta or target_parent_id in self.SPECIAL_ITEM_ID):
            raise ValueError(f'Invalid target parent ID: {target_parent_id!r}')

        if not (target_index is None or target_index >= 0):
            raise ValueError(f'Invalid target index: {target_index!r}')

        # adjust target_index
        if target_index is None:
            target_index = 0 if target_book.config['new_at_top'] else float('inf')
        target_index = min(target_index, len(target_book.toc.get(target_parent_id, ())))

        # prepare items to handle
        tasks = {}
        for current_parent_id, current_index in items:
            if not (current_index >= 0):
                raise ValueError(f'Invalid item index: {current_index!r}')

            try:
                item_id = self.toc[current_parent_id][current_index]
            except (KeyError, IndexError):
                raise ValueError(
                    f'Item not exist: {current_parent_id!r}[{current_index!r}]'
                ) from None

            if item_id not in self.meta:
                raise ValueError(f'Item not exist: {item_id!r}')

            item = (item_id, current_parent_id, current_index)

            # Prevent handling the same item again, as move_items().
            if item in tasks:
                continue

            tasks[item] = True

        # perform the tasks
        _target_index = target_index
        id_map = {}
        for item_id, _, _ in tasks:
            self._copy_item_tree(item_id, target_parent_id, _target_index, target_book,
                                 recursively, id_map)
            _target_index += 1

        return target_index, target_book_id, [v for v in id_map.values() if v]

    def _copy_item_tree(self, item_id, target_parent_id, target_index, target_book,
                        recursively, id_map):
        try:
            _item_id = id_map[item_id]
        except KeyError:
            pass
        else:
            if _item_id is not None:
                # already copied, simply link to the copy
                target_book.toc.setdefault(target_parent_id, []).insert(target_index, _item_id)
                return
            else:
                # an added copy, ignore it to prevent an infinite loop
                return

        new_item_id = self._copy_item_data(item_id, target_parent_id, target_index, target_book,
                                           id_map)

        if recursively:
            for i, child_item_id in enumerate(self.toc.get(item_id, ())):
                self._copy_item_tree(child_item_id, new_item_id, i, target_book,
                                     recursively, id_map)

    def _copy_item_data(self, item_id, target_parent_id, target_index, target_book,
                        id_map):
        try:
            item = self.meta[item_id]
        except KeyError:
            return

        # add new item and get its ID
        _item = item if item_id in target_book.meta else {**item, 'id': item_id}
        new_item_id = next(iter(target_book.add_item(_item, target_parent_id, target_index)))
        new_item = target_book.meta[new_item_id]

        # add to map
        id_map[item_id] = new_item_id
        if new_item_id != item_id and target_book.id == self.id:
            id_map[new_item_id] = None

        # copy data files
        if item.get('index'):
            if item['index'].endswith('/index.html'):
                old_index = item['index'][:-len('/index.html')]
                new_index = util.validate_filename(new_item_id)
                new_item['index'] = f'{new_index}/index.html'
            else:
                old_index = item['index']
                new_index = util.validate_filename(new_item_id) + os.path.splitext(old_index)[1]
                new_item['index'] = new_index

            old_index_file = os.path.normpath(os.path.join(self.data_dir, old_index))
            new_index_file = os.path.normpath(os.path.join(target_book.data_dir, new_index))
            if os.path.lexists(old_index_file):
                util.fs.copy(old_index_file, new_index_file)

        # copy cached favicon
        if target_book.id != self.id:
            for _ in range(1):
                old_icon_file = self.get_icon_file(item)
                if not old_icon_file:
                    break

                favicon_dir = os.path.join(self.tree_dir, 'favicon', '')
                if not os.path.normcase(old_icon_file).startswith(os.path.normcase(favicon_dir)):
                    break

                try:
                    new_base = os.path.dirname(new_index_file)
                except UnboundLocalError:
                    new_base = target_book.data_dir
                new_icon_file = os.path.join(target_book.tree_dir, 'favicon', os.path.basename(old_icon_file))
                new_item['icon'] = pathname2url(os.path.relpath(new_icon_file, new_base))
                if os.path.lexists(old_icon_file) and not os.path.lexists(new_icon_file):
                    util.fs.copy(old_icon_file, new_icon_file)

        return new_item_id

    def recycle_item(self, current_parent_id, current_index):
        """Singular version shortcut of recycle_items()."""
        return self.recycle_items(((current_parent_id, current_index),))

    def recycle_items(self, items):
        """Move items to the recycle bin and set metadata.

        Args:
            items: an iterable of tuple (current_parent_id, current_index) in
                tree order

        Returns:
            dict: ID and original parent ID of the recycled items

        Raises:
            ValueError: if the item does not exist
        """
        # prepare items to handle
        tasks = {}
        for current_parent_id, current_index in items:
            if not (current_index >= 0):
                raise ValueError(f'Invalid item index: {current_index!r}')

            try:
                item_id = self.toc[current_parent_id][current_index]
            except (KeyError, IndexError):
                raise ValueError(
                    f'Item not exist: {current_parent_id!r}[{current_index!r}]'
                ) from None

            if item_id not in self.meta:
                raise ValueError(f'Item not exist: {item_id!r}')

            item = (item_id, current_parent_id, current_index)

            # Prevent handling the same item again, as move_items().
            if item in tasks:
                continue

            tasks[item] = True

        # perform the tasks
        recycle_ts = _id_now()

        try:
            it = reversed(tasks)
        except TypeError:
            # Python 3.7 does not support dict reverse
            it = reversed(tuple(tasks))

        for _, current_parent_id, current_index in it:
            # remove from parent TOC
            del self.toc[current_parent_id][current_index]
            if not self.toc[current_parent_id]:
                del self.toc[current_parent_id]

        # handle unreachable items
        reachable_items = {}
        for root_id in self.SPECIAL_ITEM_ID:
            self.get_reachable_items(root_id, reachable_items)

        recycled = {}
        for item_id, current_parent_id, _ in tasks:
            if item_id not in reachable_items and item_id not in recycled:
                recycled[item_id] = current_parent_id
                self.meta[item_id]['parent'] = current_parent_id
                self.meta[item_id]['recycled'] = recycle_ts

        if recycled:
            target_parent_id = self.RECYCLE_ITEM_ID
            target_index = 0 if self.config['new_at_top'] else len(self.toc.get(target_parent_id, ()))
            self.toc.setdefault(target_parent_id, [])[target_index:target_index] = recycled

        return recycled

    def unrecycle_item(self, current_parent_id, current_index):
        """Singular version shortcut of unrecycle_items()."""
        return self.unrecycle_items(((current_parent_id, current_index),))

    def unrecycle_items(self, items):
        """Move items from the recycle bin to the original parent.

        Args:
            items: an iterable of tuple (current_parent_id, current_index) in
                tree order

        Returns:
            dict: ID and parent ID of the unrecycled items

        Raises:
            ValueError: if the item does not exist
        """
        # prepare items to handle
        tasks = {}
        for current_parent_id, current_index in items:
            if not (current_index >= 0):
                raise ValueError(f'Invalid item index: {current_index!r}')

            try:
                item_id = self.toc[current_parent_id][current_index]
            except (KeyError, IndexError):
                raise ValueError(
                    f'Item not exist: {current_parent_id!r}[{current_index!r}]'
                ) from None

            if item_id not in self.meta:
                raise ValueError(f'Item not exist: {item_id!r}')

            item = (item_id, current_parent_id, current_index)

            # Prevent handling the same item again, as move_items().
            if item in tasks:
                continue

            tasks[item] = True

        # perform the tasks
        try:
            it = reversed(tasks)
        except TypeError:
            # Python 3.7 does not support dict reverse
            it = reversed(tuple(tasks))

        for _, current_parent_id, current_index in it:
            # remove from parent TOC
            del self.toc[current_parent_id][current_index]
            if not self.toc[current_parent_id]:
                del self.toc[current_parent_id]

        unrecycled = {}
        for item_id, _, _ in tasks:
            # Ignore the duplicated entry in case multiple entries of item_id
            # is seen (which should generally not happen).
            if item_id in unrecycled:
                continue

            try:
                target_parent_id = self.meta[item_id].pop('parent')
            except KeyError:
                target_parent_id = self.ROOT_ITEM_ID

            try:
                del self.meta[item_id]['recycled']
            except KeyError:
                pass

            # insert to root instead if the original parent no more exists
            if not (target_parent_id in self.meta or target_parent_id in self.SPECIAL_ITEM_ID):
                target_parent_id = self.ROOT_ITEM_ID

            # record target in the map
            unrecycled[item_id] = target_parent_id

        if unrecycled:
            map_parent_items = {}
            for item_id, target_parent_id in unrecycled.items():
                map_parent_items.setdefault(target_parent_id, []).append(item_id)

            for target_parent_id, item_ids in map_parent_items.items():
                target_index = 0 if self.config['new_at_top'] else len(self.toc.get(target_parent_id, ()))
                self.toc.setdefault(target_parent_id, [])[target_index:target_index] = item_ids

        return unrecycled

    def delete_item(self, current_parent_id, current_index):
        """Singular version shortcut of delete_items()."""
        return self.delete_items(((current_parent_id, current_index),))

    def delete_items(self, items):
        """Delete items and purge their data files.

        Args:
            items: an iterable of tuple (current_parent_id, current_index) in
                tree order

        Returns:
            list: ID of the deleted items

        Raises:
            ValueError: if the item does not exist
        """
        # prepare items to handle
        tasks = {}
        for current_parent_id, current_index in items:
            if not (current_index >= 0):
                raise ValueError(f'Invalid item index: {current_index!r}')

            try:
                item_id = self.toc[current_parent_id][current_index]
            except (KeyError, IndexError):
                raise ValueError(
                    f'Item not exist: {current_parent_id!r}[{current_index!r}]'
                ) from None

            item = (item_id, current_parent_id, current_index)

            # Prevent handling the same item again, as move_items().
            if item in tasks:
                continue

            tasks[item] = True

        # perform the tasks
        old_reachable_items = {}
        for root_id in self.SPECIAL_ITEM_ID:
            self.get_reachable_items(root_id, old_reachable_items)

        try:
            it = reversed(tasks)
        except TypeError:
            # Python 3.7 does not support dict reverse
            it = reversed(tuple(tasks))

        for _, current_parent_id, current_index in it:
            # remove from parent TOC
            del self.toc[current_parent_id][current_index]
            if not self.toc[current_parent_id]:
                del self.toc[current_parent_id]

        reachable_items = {}
        for root_id in self.SPECIAL_ITEM_ID:
            self.get_reachable_items(root_id, reachable_items)

        deleted = {item_id: True for item_id in old_reachable_items if item_id not in reachable_items}
        for item_id in deleted:
            index = self.meta.get(item_id, {}).get('index')
            if index:
                if index.endswith('/index.html'):
                    index = index[:-len('/index.html')]
                entry = os.path.join(self.data_dir, index)

                # silently pass if index file does not exist
                if os.path.lexists(entry):
                    util.fs.delete(entry)

            try:
                del self.meta[item_id]
            except KeyError:
                pass

            try:
                del self.toc[item_id]
            except KeyError:
                pass

        return list(deleted)

    def sort_item(self, item_id, key=None, reverse=False, recursively=False):
        return self.sort_items((item_id,), key, reverse, recursively)

    def sort_items(self, items, key=None, reverse=False, recursively=False):
        """Sort given items.

        Args:
            items: an iterable of container item_id in tree order
            key: the key to sort, which can be 'reverse', 'type', 'title',
                'index', 'source', 'create', 'modify', or 'marked'
            recursively: also sort items in all descendant items

        Raises:
            ValueError: if the item does not exist
        """
        if recursively:
            item_ids = {}
            for item_id in items:
                for desc_item_id in self.get_reachable_items(item_id, item_ids):
                    item_ids[desc_item_id] = True
        else:
            item_ids = items

        for item_id in item_ids:
            self._sort_item(item_id, key, reverse)

    def _sort_item(self, item_id, key, reverse):
        try:
            toc = self.toc[item_id]
        except KeyError:
            # no toc to sort
            return

        if key == 'reverse':
            toc.reverse()
        elif key == 'id':
            toc.sort(reverse=reverse)
        else:
            try:
                keyfunc = getattr(self, f'_sort_items_keyfunc_{key}')
            except AttributeError:
                raise ValueError(f'Unknown sort key: {key!r}')
            toc.sort(key=keyfunc, reverse=reverse)

    _sort_items_map_type_value = {
        'folder': -1,
        'bookmark': 1,
        'postit': 2,
        'note': 3,
    }

    def _sort_items_keyfunc_type(self, item_id):
        type = self.meta.get(item_id, {}).get('type')
        try:
            value = self._sort_items_map_type_value[type]
        except KeyError:
            value = 0
        return value

    def _sort_items_keyfunc_title(self, item_id):
        return self.meta.get(item_id, {}).get('title', '')

    def _sort_items_keyfunc_index(self, item_id):
        return self.meta.get(item_id, {}).get('index', '')

    def _sort_items_keyfunc_source(self, item_id):
        return self.meta.get(item_id, {}).get('source', '')

    def _sort_items_keyfunc_create(self, item_id):
        return self.meta.get(item_id, {}).get('create', '')

    def _sort_items_keyfunc_modify(self, item_id):
        return self.meta.get(item_id, {}).get('modify', '')

    def _sort_items_keyfunc_marked(self, item_id):
        return bool(self.meta.get(item_id, {}).get('marked'))

    def load_item_postit(self, item_id):
        """Get content of a postit item.

        Args:
            item_id: ID of the postit item

        Returns:
            str: content of the postit item

        Raises:
            ValueError: if the provided item does not exist, item is not a
                postit, item index missing, etc.
        """
        try:
            item = self.meta[item_id]
        except KeyError:
            raise ValueError(f'Item not exist: {item_id!r}') from None

        if item.get('type') != 'postit':
            raise ValueError(f'Item is not a postit: {item_id!r}')

        index = item.get('index')
        if not index:
            raise ValueError(f'Item index missing: {item_id!r}')

        index_file = os.path.join(self.data_dir, index)

        if not os.path.lexists(index_file):
            return ''

        return html.unescape(self.load_postit_file(index_file))

    def save_item_postit(self, item_id, content, auto_modify=True):
        """Save content for a postit item.

        Args:
            item_id: ID of the postit item
            content: new content to save
            auto_modify: automatically update the 'modify' property

        Returns:
            dict: ID and updated meta of the saved postit item

        Raises:
            ValueError: if the provided item does not exist, item is not a
                postit, item index missing, etc.
        """
        try:
            item = self.meta[item_id]
        except KeyError:
            raise ValueError(f'Item not exist: {item_id!r}') from None

        if item.get('type') != 'postit':
            raise ValueError(f'Item is not a postit: {item_id!r}')

        index = item.get('index')
        if not index:
            raise ValueError(f'Item index missing: {item_id!r}')

        index_file = os.path.join(self.data_dir, index)

        os.makedirs(os.path.dirname(index_file), exist_ok=True)
        self.save_postit_file(index_file, html.escape(content, False))

        item['title'] = re.sub(r'\n[\s\S]*$', '', content, flags=re.M)

        if auto_modify:
            item['modify'] = _id_now()

        return {item_id: item}

    def add_item_subpage(self, item_id, title=None, ext='.html', base=None):
        """Generate a subpage for an item.

        Args:
            item_id: ID of the item
            title: title of the subpage, or None for index
            ext: file extension of the subpage ('.html' or '.md')
            base: base file (from data dir) to create the subpage, or None to
                be relative to item index

        Returns:
            str: ID of the item

        Raises:
            ValueError: if the provided item does not exist, item index
                missing, item is single-file, file already exists, etc.
        """
        try:
            item = self.meta[item_id]
        except KeyError:
            raise ValueError(f'Item not exist: {item_id!r}') from None

        index = item.get('index')
        if not index:
            raise ValueError(f'Item index missing: {item_id!r}')

        if not index.endswith('/index.html'):
            raise ValueError(f"Index page is not '*/index.html': {item_id!r}")

        item_dir = os.path.normpath(os.path.join(self.data_dir, os.path.dirname(index)))

        if base is None:
            base = index
        base = os.path.join(self.data_dir, base)

        dst = util.validate_filename(('index' if title is None else title) + ext)
        dst = os.path.normpath(os.path.join(os.path.dirname(base), dst))

        if os.path.lexists(dst):
            raise ValueError(f'Page already exists: {self.get_subpath(dst)!r}')

        try:
            tpl = self.get_template(item.get('type'), ext)
        except ValueError:
            data = ''
        else:
            data = util.format_string(tpl, {
                'NOTE_TITLE': html.escape(item.get('title', '') if title is None else title),
                'SCRAPBOOK_DIR': html.escape(util.get_relative_url(self.top_dir, dst, start_is_dir=False)),
                'TREE_DIR': html.escape(util.get_relative_url(self.tree_dir, dst, start_is_dir=False)),
                'DATA_DIR': html.escape(util.get_relative_url(self.data_dir, dst, start_is_dir=False)),
                'ITEM_DIR': html.escape(util.get_relative_url(item_dir, dst, start_is_dir=False)),
            })

        util.fs.save(dst, data.encode('UTF-8'))

        return item_id



import sys as _sys
import zipfile as _zipfile
from zipfile import *

__all__ = _zipfile.__all__


if _sys.version_info < (3, 8, 7):
    # Fix an issue causing zip not truncated
    class ZipFile(ZipFile):
        def _write_end_record(self):
            super()._write_end_record()
            if self.mode == "a":
                self.fp.truncate()
            self.fp.flush()
